function viewFiltered(InFile, Fcenter, Fwidth, Q, t1, t2)
% VIEWFILTERED  View Bandpass Filtered Quad Cell Data
%
%   Fcenter and Fwidth in Hz define a 512 point FIR filter
% that has enough resolution to extract individual lines from
% the tip/tilt jitter.  
%   Q = 1, 2 or 3 specifies which quad cell
%   t1, t2 are start and stop times of time domain calcs and plots
%
%   C. Shelton 10/4/05

load(InFile,'data');
npts = size(data,1)

%  Sampling frequency (Hz)
Fs = 500;

%  Channel assignments
x = [1 4 7];
y = [2 5 8];
s = [3 6 9];

%  Bandpass start and stop frequencies
f1 = (Fcenter - Fwidth/2)/(Fs/2);
f2 = (Fcenter + Fwidth/2)/(Fs/2);
%  Plot start and stop frequencies
fw = 2*Fwidth;
if (fw<10)
    fw = 10;
end
fa = (Fcenter - fw);
fb = (Fcenter + fw);
if (fa<0) 
    fa = 0; 
end;
if (fb>Fs/2) 
    fb = Fs/2; 
end;
fa = fa/1000; 
fb = fb/1000;

% Make an FIR bandpass filter
%  with enough points to resolve fine detail
n = 1024;
ff = fir1(n,[f1 f2]);
freqz(ff,1,512);
Hd = dfilt.dffir(ff);

% Do the filtering
fx = filter(Hd,data(:,x(Q)));
fy = filter(Hd,data(:,y(Q)));

% Compute PSDs
[Pxx,f] = pmtm(fx,4.0,2048,Fs);
psdxf = dspdata.psd(Pxx,'Fs',Fs);
[Pxx,f] = pmtm(fy,4.0,2048,Fs);
psdyf = dspdata.psd(Pxx,'Fs',Fs);
[Pxx,f] = pmtm(data(:,x(Q)),4.0,2048,Fs);
psdx = dspdata.psd(Pxx,'Fs',Fs);
[Pxx,f] = pmtm(data(:,y(Q)),4.0,2048,Fs);
psdy = dspdata.psd(Pxx,'Fs',Fs);

figure(1);
%  Power spectra plots
subplot(4,1,1);
plot(psdx)
ylim([-80,0]);
title('PSD of X')
ylabel('Power/frequency (dBmm/Hz)')

subplot(4,1,2);
plot(psdy)
ylim([-80,0]);
title('PSD of Y')
ylabel('Power/frequency (dBmm/Hz)')

subplot(4,1,3);
plot(psdxf)
ylim([-80,0]);
xlim([fa, fb]);
title('PSD of Filtered X')
ylabel('Power/frequency (dBmm/Hz)')

subplot(4,1,4);
plot(psdyf)
ylim([-80,0]);
xlim([fa, fb]);
title('PSD of Filtered Y')
ylabel('Power/frequency (dBmm/Hz)')

figure(2);
%  Time domain plots and calcs
%    t1, t2 = start time, stop time (sec)
a = round(Fs*t1) + 1;
b = round(Fs*t2) + 1;
if b<=a
    b = a + 0.1;
end
if (b>npts)
    b=npts;
end
t = 1:npts;
t = t/Fs;

subplot(4,1,1);
%  Plot filtered X
plot(t(a:b), fx(a:b))
ylim([-0.3,0.3]);
title('Filtered X')
xlabel('Time (sec)')
ylabel('Position (mm)')

subplot(4,1,2);
%  Plot filtered Y
plot(t(a:b), fy(a:b))
ylim([-0.3,0.3]);
title('Filtered Y')
xlabel('Time (sec)')
ylabel('Position (mm)')

subplot(4,1,3);
%  Plot unfiltered Y
plot(t(a:b), data(a:b,8))
%ylim([-1,1]);
title('Unfiltered Y')
xlabel('Time (sec)')
ylabel('Position (mm)')

subplot(4,1,4);
%  Find and plot coeficients of best fit 
%    y = ax + b for each n second segment
%    n = number of seconds per segment
n = 5;
n = n*Fs;
w = hann(n);
j = 0;
for i = a:500:b-n-1;
    j = j+1;
    t(j) = i/Fs;
    p = polyfit(w.*fx(i:i+n-1),w.*fy(i:i+n-1),1);
    pfit(j) = p(1);
end
plot(t(1:j), pfit(1:j))
%ylim([-1,1]);
title('Best Fit y/x')
xlabel('Time (sec)')
ylabel('Slope (y/x)')
grid on

figure(3);
%  Correlation plots
subplot(2,1,1);
%  Plot X, Y crosscorrelation
c = xcorr(fy(a:b), fx(a:b));
%  d = half-width of plot (sec)
d = 0.05;
d = round(Fs*d);
if d>(b-a)/2
    d = (b-a)/2;
end
t2 = -d:d;
t2 = t2/Fs;
aa = (b-a+1) - d;
bb = (b-a+1) + d;
plot(t2, c(aa:bb))
title('XCORR(X, Y)')
xlabel('Time (sec)')
ylabel('Raw crosscorrelation')

subplot(2,1,2);
%  Plot X vs Y
plot(fx(a:b), fy(a:b))
xlim([-0.3,0.3]);
ylim([-0.3,0.3]);
title('X vs Y')
xlabel('X Position (mm)')
ylabel('Y Position (mm)')
grid on


