/*****************************************************************************/
/*                    Buffered      analog output example                    */
/*                                                                           */
/*  This example shows how to use the powerdaq API to perform a buffered     */  
/*  update. The update is performed in a hardware timed fashion              */
/*  that is appropriate for fast speed generation (> 1MHz).                  */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/*      Copyright (C) 2001 United Electronic Industries, Inc.                */
/*      All rights reserved.                                                 */
/*---------------------------------------------------------------------------*/
/*                                                                           */ 
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include "win_sdk_types.h"
#include "powerdaq.h"
#include "powerdaq32.h"

#include "ParseParams.h"


typedef enum _state
{
   closed,
   unconfigured,
   configured,
   running
} tState;

typedef struct _bufferedAoData
{
   int abort;                    // set to TRUE to abort generation
   int board;                    // board number to be used for the AI operation
   int handle;                   // board handle
   int adapterType;              // adapter type, shoul be atMF or atPD2AO
   int nbOfChannels;             // number of channels
   unsigned long channelList[64];
   int nbOfPointsPerChannel;     // number of samples per channel
   int nbOfFrames;
   int halfFIFOSize;             // Half size of the DAC FIFO
   double updateRate;            // sampling frequency on each channel
   tState state;                 // state of the acquisition session
   unsigned int *rawBuffer;      // address of the buffer allocated by the driver to store
} tBufferedAoData;


int InitBufferedAO(tBufferedAoData *pAoData);
int BufferedAO(tBufferedAoData *pAoData, double *buffer);
void CleanUpBufferedAO(tBufferedAoData *pAoData);

double getWave(int index, int channel);


static tBufferedAoData G_AoData;

// exit handler
void BufferedAOExitHandler(int status, void *arg)
{
   CleanUpBufferedAO((tBufferedAoData *)arg);
}


int InitBufferedAO(tBufferedAoData *pAoData)
{
   int retVal = 0;
   Adapter_Info adaptInfo;

   // get adapter type
   retVal = __PdGetAdapterInfo(pAoData->board, &adaptInfo);
   if (retVal < 0)
   {
      printf("BufferedAO: __PdGetAdapterInfo error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   pAoData->adapterType = adaptInfo.atType;
   if(pAoData->adapterType & atMF)
      printf("This is an MFx board\n");
   else if(pAoData->adapterType & atPD2AO)
      printf("This is an AO32 board\n");
   else
   {
      printf("This device doesn't have AO capability\n");
      exit(EXIT_FAILURE);
   }

   pAoData->handle = PdAcquireSubsystem(pAoData->board, AnalogOut, 1);
   if(pAoData->handle < 0)
   {
      printf("BufferedAO: PdAcquireSubsystem failed\n");
      exit(EXIT_FAILURE);
   }

   pAoData->state = unconfigured;

   retVal = _PdAOutReset(pAoData->handle);
   if (retVal < 0)
   {
      printf("BufferedAO: PdAOutReset error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   // need also to call this function if the board is a PD2-AO-xx
   if(pAoData->adapterType & atPD2AO)
   {
      retVal = _PdAO32Reset(pAoData->handle);
      if (retVal < 0)
      {
         printf("BufferedAO: PdAAO32Reset error %d\n", retVal);
         exit(EXIT_FAILURE);
      }
   }

   return 0;
}


int BufferedAO(tBufferedAoData *pAoData, double *voltBuffer)
{
   int retVal;
   DWORD aoCfg;
   DWORD cvDivider;
   DWORD count = 0;
   DWORD eventsToNotify = eTimeout | eFrameDone | eBufferDone | eBufferError | eStopped;
   DWORD events;
   DWORD multiplier;
   DWORD scanIndex, numScans, samplesPerFrame;
   int i, j;
   

   // setup the board to use hardware internal clock
   aoCfg = AOB_CVSTART0;  // 11 MHz internal timebase
//   aoCfg |= AOB_REGENERATE; // This configuration bits causes the board to automatically
                            // regenerate from the data stored in this FIFO.
      
   retVal = _PdRegisterBuffer(pAoData->handle, (WORD**)&pAoData->rawBuffer, AnalogOut,
                              pAoData->nbOfFrames, pAoData->nbOfPointsPerChannel,
                              pAoData->nbOfChannels, 
                              BUF_BUFFERRECYCLED | BUF_BUFFERWRAPPED | BUF_DWORDVALUES);
   if (retVal < 0)
   {
      printf("BufferedAO: PdRegisterBuffer error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   // write data to generate in the buffer
   // Convert volts to binary
   /*retVal = PdAOutVoltsToRaw(pAoData->board, aoCfg, voltBuffer, pAoData->rawBuffer, 
                             pAoData->nbOfPointsPerChannel * 
                             pAoData->nbOfChannels);
   if(retVal < 0)
   {
      printf("BufferedAO: PdAOutVoltsToRaw failed with code %x\n", retVal);
      exit(EXIT_FAILURE);
   } */
   
   if(pAoData->adapterType & atPD2AO)
      multiplier = 0xFFFF;
   else
      multiplier = 0xFFF;
   
   samplesPerFrame = pAoData->nbOfPointsPerChannel * pAoData->nbOfChannels;
   for(i=0; i<pAoData->nbOfFrames; i++)
      for(j=0; j<samplesPerFrame; j++)
         pAoData->rawBuffer[i*samplesPerFrame+j] = (voltBuffer[j] + 10.0) / 20.0 * multiplier;


   // set clock divider, assuming that we use the 11MHz timebase
   cvDivider = 11000000.0 / pAoData->updateRate -1;
   
   if(pAoData->adapterType & atMF)
      retVal = _PdAOutAsyncInit(pAoData->handle, aoCfg, 
                               cvDivider, eventsToNotify);
   else
      retVal = _PdAOAsyncInit(pAoData->handle, aoCfg, 
                               cvDivider, eventsToNotify, 
                               pAoData->nbOfChannels, (DWORD*)pAoData->channelList);
   if (retVal < 0)
   {
      printf("BufferedAO: PdAOutAsyncInit error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   pAoData->state = configured;

   retVal = _PdSetUserEvents(pAoData->handle, AnalogOut, eventsToNotify);
   if (retVal < 0)
   {
      printf("BufferedAO: _PdSetUserEvents error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   if(pAoData->adapterType & atMF)
      retVal = _PdAOutAsyncStart(pAoData->handle);
   else
      retVal = _PdAOAsyncStart(pAoData->handle);
   if (retVal < 0)
   {
      printf("BufferedAO: PdAOutAsyncStart error %d\n", retVal);
      exit(EXIT_FAILURE);
   }

   pAoData->state = running;

   while (!pAoData->abort)
   {
      if((aoCfg & AOB_REGENERATE) == 0)
      {
          events = _PdWaitForEvent(pAoData->handle, eventsToNotify, 5000);
    
          printf("Received events 0x%x\n", events);
          
          retVal = _PdSetUserEvents(pAoData->handle, AnalogOut, eventsToNotify);
          if (retVal < 0)
          {
             printf("BufferedAO: _PdSetUserEvents error %d\n", retVal);
             exit(EXIT_FAILURE);
          }
       
          if (events & eTimeout)
          {
             printf("BufferedAO: wait timed out\n"); 
             exit(EXIT_FAILURE);
          }
       
          if ((events & eBufferError) || (events & eStopped))
          {
             printf("BufferedAO: buffer error\n");
             exit(EXIT_FAILURE);
          }
       
          if ((events & eBufferDone) || (events & eFrameDone))
          {
             if(pAoData->adapterType & atMF)
                retVal = _PdAOutGetBufState(pAoData->handle, pAoData->nbOfPointsPerChannel, 
                                         SCANRETMODE_MMAP, &scanIndex, &numScans);
             else
                retVal = _PdAOGetBufState(pAoData->handle, pAoData->nbOfPointsPerChannel, 
                                         SCANRETMODE_MMAP, &scanIndex, &numScans);
             if(retVal < 0)
             {
                printf("BufferedAO: buffer error\n");
                exit(EXIT_FAILURE);
             }
    
                 
             count = count + numScans;
       
             printf("BufferedAO: sent %d scans at %d\n", numScans, scanIndex);
          }
      }
      else
      {
          printf("Regenerating...\n");
          sleep(1);
      }
   }

   return retVal;
}


void CleanUpBufferedAO(tBufferedAoData *pAoData)
{
   int retVal;
      
   if(pAoData->state == running)
   {
      if(pAoData->adapterType & atMF)
         retVal = _PdAOutAsyncStop(pAoData->handle);
      else
         retVal = _PdAOAsyncStop(pAoData->handle);
      if (retVal < 0)
         printf("BufferedAO: PdAOutAsyncStop error %d\n", retVal);

      pAoData->state = configured;
   }

   if(pAoData->state == configured)
   {
      retVal = _PdClearUserEvents(pAoData->handle, AnalogOut, eAllEvents);
      if (retVal < 0)
         printf("BufferedAI: PdClearUserEvents error %d\n", retVal);
   
      if(pAoData->adapterType & atMF)
         retVal = _PdAOutAsyncTerm(pAoData->handle); 
      else
         retVal = _PdAOAsyncTerm(pAoData->handle);
      if (retVal < 0)
         printf("BufferedAO: PdAOutAsyncTerm error %d\n", retVal);
   
      retVal = _PdUnregisterBuffer(pAoData->handle, (WORD *)pAoData->rawBuffer, AnalogOut);
      if (retVal < 0)
         printf("BufferedAO: PdUnregisterBuffer error %d\n", retVal);

      pAoData->state = unconfigured;
   }

   if(pAoData->handle > 0 && pAoData->state == unconfigured)
   {
      retVal = PdAcquireSubsystem(pAoData->handle, AnalogOut, 0);
      if (retVal < 0)
         printf("BufferedAO: PdReleaseSubsystem error %d\n", retVal);
   }

   pAoData->state = closed;
}


//////////////////////////////////////////////////////////////////////////
// Function to react to SIGINT (Ctrl+C)
void SigInt(int signum)
{
   if (signum == SIGINT)
   {
      printf("Ctrl+C detected. generation stopped\n");
      G_AoData.abort = TRUE;
   }
}


double getWave(int index, int channel)
{
   int nbOfPointsPerCycle;

   nbOfPointsPerCycle = ((index / 500) + 1) + 500;
   
   return(10.0 * sin(2.0 * 3.1415 * index /** (channel + 1)*/ / nbOfPointsPerCycle));
}


int main(int argc, char *argv[])
{
   double *voltBuffer = NULL;
   int nbOfSubBuffers;
   int i, j;
   PD_PARAMS params = {0, 1, {0}, 1000.0, 0};
   
   ParseParameters(argc, argv, &params);

   // initializes acquisition session parameters
   G_AoData.board = params.board;
   G_AoData.nbOfChannels = params.nb_channels;
   for(i=0; i<params.nb_channels; i++)
       G_AoData.channelList[i] = params.channels[i];
   G_AoData.handle = 0;
   G_AoData.nbOfPointsPerChannel = 2000;
   G_AoData.updateRate = params.frequency;
   G_AoData.nbOfFrames = 4;
   G_AoData.halfFIFOSize = 1024;
   G_AoData.state = closed;
   G_AoData.abort = FALSE;
   G_AoData.rawBuffer = NULL;

   // setup exit handler that will clean-up the acquisition session
   // if an error occurs
   on_exit(BufferedAOExitHandler, &G_AoData);

   // set up SIGINT handler
   signal(SIGINT, SigInt);

   // allocate memory for the generation buffer
   voltBuffer = (double *) malloc(G_AoData.nbOfChannels * G_AoData.nbOfPointsPerChannel *
                              sizeof(double));
   if(voltBuffer == NULL)
   {
      printf("BufferedAO: could not allocate enough memory for the generation buffer\n");
      exit(EXIT_FAILURE);
   }

   // writes data to buffer. Each channel will generate a sinus of different frequency
   nbOfSubBuffers = G_AoData.nbOfPointsPerChannel * G_AoData.nbOfChannels / G_AoData.halfFIFOSize;
   for(i=0; i<G_AoData.nbOfPointsPerChannel; i++)
   {
      for(j=0; j<G_AoData.nbOfChannels; j++)
      {
         voltBuffer[i*G_AoData.nbOfChannels + j] = getWave(i, j);
      }
   }

   // initializes acquisition session
   InitBufferedAO(&G_AoData);

   // run the acquisition
   BufferedAO(&G_AoData, voltBuffer);

   // Cleanup acquisition
   CleanUpBufferedAO(&G_AoData);

   // free acquisition buffer
   free(voltBuffer);
   
   return 0;
}
