unit AnInp;

interface

uses
{$IFDEF MSWINDOWS}
   Windows, Forms, ExtCtrls, Graphics,
{$ENDIF}
{$IFDEF LINUX}
   QForms, QExtCtrls, QGraphics,
{$ENDIF}
   Classes, SysUtils, Types,
   DAQDefs, VisualThread, pwrdaq, pdfw_def,
{$IFNDEF DemoModule}
   pwrdaq32, pwrdaq_struct, pwrdaq_LinuxAPI,
{$ELSE}
   PD32Interface,
{$ENDIF}
   pd_hcaps;

type

  TAcquisitionSetup = record
    // for hardware setup
    Adapter: THandle;              // adapter handle
    BoardType: DWORD;              // type of the PowerDAQ board
    Resolution: DWORD;             // adapter resolution
    FIFOSize: DWORD;               // FIFO size
    Frequency: DWORD;              // acquisition Frequency
    Channel: DWORD;                // active channel
    NumChannels: DWORD;            // total number of channel to acquisition
    DefGain: DWORD;                // default gain (for all channels)
    AInMode: DWORD;                // Single-ended/Differential
    AInType: DWORD;                // Unipolar/Bipolar
    AInRange: DWORD;               // Low/High
    // for visualisation setup
    PaintRect: TRect;
    PaintCanvas: TCanvas;
  end;

  // base acquisition thread
  TAnalogInput = class(TThread)
  private
    AcqSetup: TAcquisitionSetup;
    dwError: DWORD;
    hNotifyEvent : THandle;
    FActiveChannel: Integer;
    AInChList: array [0..63] of DWORD;
    AInBuffer: PAnalogInputBuffer;
    dwAInScanSize: DWORD;
    dwAInScans: DWORD;
    Visualisation: TVisualThread;
    procedure SetActiveChannel(Value: Integer);
  protected
    procedure DoTerminate; override;
  public
    constructor Create(AAcqSetup: TAcquisitionSetup);
    procedure Execute; override;
    procedure DoGetData; virtual;
    property ActiveChannel: Integer read FActiveChannel write SetActiveChannel;
  end;

implementation

// methods for AcquisitionThread
constructor TAnalogInput.Create(AAcqSetup: TAcquisitionSetup);
begin
  // assign setup record
  AcqSetup := AAcqSetup;
  // create visualisation thread
  Visualisation := TVisualThread.Create(AcqSetup.PaintCanvas, AcqSetup.PaintRect);
  inherited Create(False);
  FreeOnTerminate := True;
end;

procedure TAnalogInput.Execute;
var
  i: Integer;
  dwAInCfg,
  dwConvType,
  dwAInPreTrigCount,
  dwAInPostTrigCount,
  dwAInCvClkDiv,
  dwAInClClkDiv,
  dwEventsNotify,
  dwEventsReceived,
  dwAInFrames,
  dwWaitTime : DWORD;
begin
  // analog input - init sequence
  try
    if (AcqSetup.BoardType = atPD2MF) or (AcqSetup.BoardType = atPDMF)
      then dwConvType :=  AIB_CLSTART0 or AIB_CLSTART1 or  AIB_CVSTART0
      else dwConvType :=  AIB_CLSTART0 or AIB_CVSTART0 or  AIB_CVSTART1;

    dwAInCfg := AcqSetup.AInMode                      // Modes (Single ended/Differential)
             or AcqSetup.AInType                      // Input Type (Unipolar/Bipolar)
             or AcqSetup.AInRange                     // Input Range (Low/High)
             or dwConvType
             or AIB_INTCLSBASE                        //
             or AIB_INTCVSBASE;                       // Select 33 MHz Source

    // build Channel list
    for i:=0 to AcqSetup.NumChannels-1 do
      AInChList [i] := i + (AcqSetup.DefGain SHL 6);

    // init counter variables
    dwAInPreTrigCount := 0;
    dwAInPostTrigCount := 0;
    dwAInCvClkDiv := Round(BaseFrequency[ LongBool(dwAInCfg and AIB_INTCLSBASE)]  / (AcqSetup.Frequency)) - 1;

    dwAInClClkDiv := dwAInCvClkDiv;

    // init buffer variables
    dwAInScanSize := AcqSetup.NumChannels;
    dwAInScans := Round( AcqSetup.Frequency / (10 * dwAInScanSize));
    if dwAInScans < (AcqSetup.PaintRect.Right-AcqSetup.PaintRect.Left) * 2
      then dwAInScans := (AcqSetup.PaintRect.Right-AcqSetup.PaintRect.Left) * 2;
    dwAInFrames   := 16;
    dwWaitTime    := Round(AcqSetup.FIFOSize / (AcqSetup.Frequency * dwAInScanSize) * 750);
    if dwWaitTime < 3000 then dwWaitTime := 3000;

    // get subsystem in use
{$IFDEF MSWINDOWS}
    if not PdAdapterAcquireSubsystem(AcqSetup.Adapter, @dwError, AnalogIn, 1)
      then raise TPwrDaqException.Create('PdAdapterAcquireSubsystem', dwError);

    // get buffer in use
    if not _PdAllocateBuffer(@AInBuffer, dwAInFrames, dwAInScans, dwAInScanSize, @dwError)
      then raise TPwrDaqException.Create('_PdAllocateBuffer', dwError);

    // connect buffer to AIn subsystem
    if not _PdRegisterBuffer(AcqSetup.Adapter, @dwError, PWORD(AInBuffer), AnalogIn, AIB_BUFFERWRAPPED or AIB_BUFFERRECYCLED)
      then raise TPwrDaqException.Create('_PdRegisterBuffer', dwError);
{$ENDIF}
{$IFDEF LINUX}
    AcqSetup.Adapter := PdAcquireSubsystem(AcqSetup.Adapter, AnalogIn, 1);
    if (AcqSetup.Adapter < 0)
      then raise TPwrDaqException.Create('PdAdapterAcquireSubsystem', dwError);

    // connect buffer to AIn subsystem
    dwError := _PdRegisterBuffer(AcqSetup.Adapter, @AInBuffer, AnalogIn, dwAInFrames, dwAInScans, dwAInScanSize, AIB_BUFFERWRAPPED or AIB_BUFFERRECYCLED);
    if (dwError < 0)
      then raise TPwrDaqException.Create('_PdRegisterBuffer', dwError);
{$ENDIF}

    // set notification events
    dwEventsNotify := eFrameDone or eBufferDone or eBufferWrapped or
                      eTimeout or eBufferError or eStopped; // or eConvError or eScanError;

    // configure AIn subsystem.
{$IFDEF MSWINDOWS}
    if not _PdAInAsyncInit(AcqSetup.Adapter, @dwError,
                           dwAInCfg,
{$IFDEF WithFIFOSIZE}      AcqSetup.FIFOSize div 2,  {$ENDIF}     // < New! FIFO size
                           dwAInPreTrigCount,
                           dwAInPostTrigCount,
                           dwAInCvClkDiv,
                           dwAInClClkDiv,
                           dwEventsNotify,
                           dwAInScanSize,
                           @AInChList)
      then raise TPwrDaqException.Create('_PdAInAsyncInit', dwError);

    // enable A/D conversions.
    if not _PdAInAsyncStart(AcqSetup.Adapter, @dwError)
      then raise TPwrDaqException.Create('_PdAInAsyncStart', dwError);

    // set private event
    if not _PdAInSetPrivateEvent(AcqSetup.Adapter, @hNotifyEvent)
      then raise TPwrDaqException.Create('_PdSetPrivateEvent', -1);
{$ENDIF}
{$IFDEF LINUX}
    dwError := _PdAInAsyncInit(AcqSetup.Adapter,
                           dwAInCfg,
                           dwAInPreTrigCount,
                           dwAInPostTrigCount,
                           dwAInCvClkDiv,
                           dwAInClClkDiv,
                           dwEventsNotify,
                           dwAInScanSize,
                           @AInChList);
    if(dwError < 0)
      then raise TPwrDaqException.Create('_PdAInAsyncInit', dwError);

    // enable A/D conversions.
    dwError := _PdAInAsyncStart(AcqSetup.Adapter);
    if(dwError < 0)
      then raise TPwrDaqException.Create('_PdAInAsyncStart', dwError);

{$ENDIF}

    // run visualisation
    if Visualisation <> nil then
    begin
      Visualisation.Channel := ActiveChannel;
      Visualisation.ScanSize := dwAInScanSize;                   // number of channels
      Visualisation.NumChannels := AcqSetup.NumChannels;
      Visualisation.FrameSize := dwAInScans * dwAInScanSize;     // size of frame in samples
      Visualisation.Buffer := AInBuffer;
      Visualisation.Offset := 0;

      // select board type (12, 14 or 16 bits)
      if AcqSetup.Resolution = 12 then
      begin
        Visualisation.XorMask := $800;
        Visualisation.AndMask := $FFF;
        Visualisation.Shift := 5;
      end
      else begin
        Visualisation.XorMask := $8000;
        Visualisation.AndMask := $FFFF;
        Visualisation.Shift := 9;
      end;
      Visualisation.Resume;
    end;

    // thread loop
    while not Terminated do
    try
{$IFDEF MSWINDOWS}
      if WaitForSingleObject(hNotifyEvent, dwWaitTime) = WAIT_OBJECT_0
{$ENDIF}
{$IFDEF LINUX}
      dwEventsReceived := _PdWaitForEvent(AcqSetup.Adapter, dwEventsNotify, dwWaitTime);
      if(dwEventsREceived and eTimeout) = 0 
{$ENDIF}
        then Synchronize(DoGetData)
        else raise TPwrDaqException.Create('Time-out detected. Acquisition stopped', 0);
    except
      raise;
    end;
  except
    Application.HandleException(Self);
  end;
end;

procedure TAnalogInput.DoGetData;
var
  dwAInEvents: Cardinal;
  ScanIndex, NumValidScans: DWORD;
begin
  try
{$IFDEF MSWINDOWS}
    if not _PdGetUserEvents(AcqSetup.Adapter, @dwError, AnalogIn, @dwAInEvents)
{$ENDIF}
{$IFDEF LINUX}
    dwError := _PdGetUserEvents(AcqSetup.Adapter, AnalogIn, @dwAInEvents);
    if(dwError < 0)
{$ENDIF}
      then raise TPwrDaqException.Create('_PdGetUserEvents', dwError);

    // if aq. stopped, test on any error
    if LongBool(eStopped and dwAInEvents) then
    begin
      // aq. stopped test
      if LongBool(eBufferError and dwAInEvents)
        then raise TPwrDaqException.Create('Buffer error detected', 0);

      // time-out test
      if LongBool(eTimeout and dwAInEvents)
        then raise TPwrDaqException.Create('Time-out detected', 0);
    end;

    // get scans using PdAInGetScans.
{$IFDEF MSWINDOWS}
    if not _PdAInGetScans(AcqSetup.Adapter, @dwError, dwAInScans, @ScanIndex, @NumValidScans)
{$ENDIF}
{$IFDEF LINUX}
    dwError := _PdAInGetScans(AcqSetup.Adapter, dwAInScans, AIN_SCANRETMODE_MMAP, @ScanIndex, @NumValidScans);
    if(dwError < 0)
{$ENDIF}
      then raise TPwrDaqException.Create('_PdAInGetScans', dwError);

    // if data available, draw it
    if LongBool(eFrameDone and dwAInEvents) and (Visualisation <> nil)
      then Visualisation.Offset := ScanIndex;       // store buffer offset (in scans, not in samples!)

  finally
    // make sure, that we're  interested in this two events,
    dwAInEvents := dwAInEvents and (not (eDataAvailable or eScanDone));

    // re-enable user events that asserted.
{$IFDEF MSWINDOWS}
    if not _PdSetUserEvents(AcqSetup.Adapter, @dwError, AnalogIn, dwAInEvents)
{$ENDIF}
{$IFDEF LINUX}
    dwError := _PdSetUserEvents(AcqSetup.Adapter, AnalogIn, dwAInEvents);
    if(dwError < 0)
{$ENDIF}
      then raise TPwrDaqException.Create('_PdClearUserEvents', dwError);
  end;
end;

procedure TAnalogInput.DoTerminate;
begin
  // stop acquisition sequence
  try

   // terminate visualisation thread
   if Visualisation <> nil then
   begin
     Visualisation.Terminate;
     //Visualisation.Free;
     //Visualisation := nil;
   end;

{$IFDEF MSWINDOWS}
   // disable A/D conversions
   if not _PdAInClearPrivateEvent(AcqSetup.Adapter, @hNotifyEvent)
     then raise TPwrDaqException.Create('_PdAInClearPrivateEvent', -1);

   // stop acquisition
   if not _PdAInAsyncStop(AcqSetup.Adapter, @dwError)
     then raise TPwrDaqException.Create('_PdAInAsyncStop', dwError);

   // clear all events
   if not _PdClearUserEvents(AcqSetup.Adapter, @dwError, AnalogIn, eAllEvents)
     then raise TPwrDaqException.Create('_PdClearUserEvents', dwError);

   // terminate acquisition
   if not _PdAInAsyncTerm(AcqSetup.Adapter, @dwError)
     then raise TPwrDaqException.Create('_PdAInAsyncTerm', dwError);

   // unregister data buffer &
   if not _PdUnregisterBuffer(AcqSetup.Adapter, @dwError, PWORD(AInBuffer))
     then raise TPwrDaqException.Create('_PdUnregisterBuffer', dwError);

   // dispose it
   if not _PdFreeBuffer(PWORD(AInBuffer), @dwError)
     then raise TPwrDaqException.Create('_PdFreeBuffer', dwError);

   // free AI subsystem
   if not PdAdapterAcquireSubsystem(AcqSetup.Adapter, @dwError, AnalogIn, 0)
     then raise TPwrDaqException.Create('PdAdapterAcquireSubsystem', dwError);
{$ENDIF}
{$IFDEF LINUX}
   // stop acquisition
   dwError := _PdAInAsyncStop(AcqSetup.Adapter);
   if(dwError < 0)
     then raise TPwrDaqException.Create('_PdAInAsyncStop', dwError);

   // clear all events
   dwError := _PdClearUserEvents(AcqSetup.Adapter, AnalogIn, eAllEvents);
   if(dwError < 0)
     then raise TPwrDaqException.Create('_PdClearUserEvents', dwError);

   // terminate acquisition
   dwError := _PdAInAsyncTerm(AcqSetup.Adapter);
   if(dwError < 0)
     then raise TPwrDaqException.Create('_PdAInAsyncTerm', dwError);

   // unregister data buffer &
   dwError := _PdUnregisterBuffer(AcqSetup.Adapter, PWORD(AInBuffer), AnalogIn);
   if(dwError < 0)
     then raise TPwrDaqException.Create('_PdUnregisterBuffer', dwError);

   // free AI subsystem
   dwError := PdAcquireSubsystem(AcqSetup.Adapter, AnalogIn, 0);
   if(dwError < 0)
     then raise TPwrDaqException.Create('PdAdapterAcquireSubsystem', dwError);
{$ENDIF}

  except
    // handle any exception
    Application.HandleException(Self);
  end;
  inherited;
end;

procedure TAnalogInput.SetActiveChannel(Value: Integer);
begin
  if FActiveChannel <> Value then
  begin
    FActiveChannel := Value;
    if Visualisation <> nil then Visualisation.Channel := Value;
  end;
end;

end.
