//===========================================================================
//
// NAME:    powerdaq.c
//
// DESCRIPTION:
//
//          PowerDAQ Linux driver
//
// AUTHOR: Frederic Villeneuve 
//         Alex Ivchenko and Sebastian Kuzminski
//
//---------------------------------------------------------------------------
//      Copyright (C) 2000,2004 United Electronic Industries, Inc.
//      All rights reserved.
//---------------------------------------------------------------------------
// For more informations on using and distributing this software, please see
// the accompanying "LICENSE" file.
//
// "Include" map
//
//  powerdaq.c
//     |
//     + pdfw_def.h
//     + powerdaq-internal.h
//     |   |
//     |   + pdfw_def.h
//     |   + win_ddk_types.h
//     |
//     + powerdaq.h
//

#include <linux/config.h>

// Turn on module versioning if the kernel was configured
// to use it and we don't want debug messages. If we want
// debug messages, versioning need to be turned off because
// the binary module pdfw_lib.o will need to use printk.
/*#if (defined(CONFIG_MODVERSIONS) && !defined(PD_DEBUG))
   #define MODVERSIONS
   #include <linux/modversions.h>
#endif*/ 

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
   #include <linux/slab.h>
#else
   #include <linux/malloc.h>
#endif
#include <linux/delay.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/workqueue.h>
#else
#include <linux/tqueue.h>
#endif
#include <linux/proc_fs.h>
#include <linux/fs.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/page.h>

#ifdef _PD_RTLPRO
   #include <rtl.h>
   #include <rtl_errno.h>
   #include <pthread.h>
   #include <rtl_time.h>
#endif

#ifdef _PD_RTL
   #include <rtl_mutex.h>
   #include <rtl_spinlock.h>
   #include <pthread.h>
   #include <time.h>
#endif

#ifdef _PD_RTAI
   #include <config.h>
   #include <rtai.h>
   #include <rtai_sched.h>
#endif


#include "./include/kernel_compat.h"

#include "./include/win_ddk_types.h"
#include "./include/pdfw_def.h"
#include "./include/powerdaq-internal.h"
#include "./include/powerdaq-extension.h"
#include "./include/powerdaq.h"
#include "./include/pdfw_if.h"
#include "./include/kvmem.h"

#include "kvmem.c" // include part of mbuff driver by Tomasz Motylewski

const char *pd_devices_by_minor[] = {
   "AIn",
   "AOut",
   "DIn",
   "DOut",
   "UCT",
   "Drv",
   "DSPCT"
};

// sources are split into firmware library and driver itself
//
#include "./include/pd_boards.h"

extern pd_board_t pd_board[PD_MAX_BOARDS];
extern int num_pd_boards;
#if defined(_PD_RTL) || defined(_PD_RTAI) || defined(_PD_RTLPRO)
pd_board_ext_t pd_board_ext[PD_MAX_BOARDS];
#define GET_BOARD_EXT(X)  (pd_board_ext_t*)(pd_board[X].extension)
#endif


// global variables for deferred processing of interrupts
// Under Linux we use bottom-halves
// Under RT-Linux we use a thread that wakes-up after the ISR is called
// Under RTAI we use a task that waits on a semaphore unlocked from the ISR
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
pthread_t       rt_isr_thread[PD_MAX_BOARDS];
#elif defined(_PD_RTAI)
RT_TASK         rt_bh_task[PD_MAX_BOARDS];
SEM             rt_bh_sem[PD_MAX_BOARDS];
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
struct work_struct pd_bh_task[PD_MAX_BOARDS];
#else
struct tq_struct pd_bh_task[PD_MAX_BOARDS];
#endif
#endif

static unsigned long intcnt;

int xferMode = 1;
int pd_major = PD_MAJOR;
int rqstirq = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
module_param(xferMode, int, 1);
module_param(pd_major, int, 1);
module_param(rqstirq, int, 1);
MODULE_ALIAS_CHARDEV_MAJOR(PD_MAJOR);
#else
MODULE_PARM(xferMode,"i");
MODULE_PARM(pd_major,"i");
MODULE_PARM(rqstirq,"i");
#endif

// Spinlocks to make driver SMP-safe
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   pthread_spinlock_t pd_fw_lock = PTHREAD_SPINLOCK_INITIALIZER;

   #define _fw_spinlock pthread_spin_lock(&pd_fw_lock);
   #define _fw_spinunlock pthread_spin_unlock(&pd_fw_lock);
#elif defined(_PD_RTAI)
   spinlock_t pd_fw_lock = SPIN_LOCK_UNLOCKED;
   unsigned long pd_fw_flags;

   #define _fw_spinlock pd_fw_flags = rt_spin_lock_irqsave(&pd_fw_lock);
   #define _fw_spinunlock rt_spin_unlock_irqrestore(pd_fw_flags, &pd_fw_lock);
#else
   spinlock_t pd_fw_lock = SPIN_LOCK_UNLOCKED;
   unsigned long pd_fw_flags;

   #define _fw_spinlock spin_lock_irqsave(&pd_fw_lock, pd_fw_flags);
   #define _fw_spinunlock spin_unlock_irqrestore(&pd_fw_lock, pd_fw_flags);
#endif

#define DIRECTEVENTS 1

// --------- test code --------
wait_queue_head_t whead;
// --------- test code ends ---

// kernel function wrappers (see implementation below)
extern void pd_kill_fasync(struct fasync_struct *fa, int sig);

extern unsigned long pd_copy_from_user32(u32* to, u32* from, u32 len);
extern unsigned long pd___copy_from_user32(u32* to, u32* from, u32 len);
extern unsigned long pd_copy_to_user32(u32* to, u32* from, u32 len);
extern unsigned long pd___copy_to_user32(u32* to, u32* from, u32 len);

extern unsigned long pd_copy_from_user16(u16* to, u16* from, u32 len);
extern unsigned long pd___copy_from_user16(u16* to, u16* from, u32 len);
extern unsigned long pd_copy_to_user16(u16* to, u16* from, u32 len);
extern unsigned long pd_copy_to_userbuf16(u16* to, u16* from, u32 len);
extern unsigned long pd___copy_to_user16(u16* to, u16* from, u32 len);

extern unsigned long pd_copy_from_user8(u8* to, u8* from, u32 len);
extern unsigned long pd___copy_from_user8(u8* to, u8* from, u32 len);
extern unsigned long pd_copy_to_user8(u8* to, u8* from, u32 len);
extern unsigned long pd___copy_to_user8(u8* to, u8* from, u32 len);

extern int pd_get_user32(u32* value, u32* address);
extern int pd___get_user32(u32* value, u32* address);
extern int pd_get_user16(u16* value, u16* address);
extern int pd___get_user16(u16* value, u16* address);

extern void pd_udelay(u32 usecs);
extern void pd_mdelay(u32 msecs);
extern void* pd_alloc_bigbuf(u32 size);
extern void pd_free_bigbuf(void* mem, u32 size);
        
int pd_sleep_on_event(int board, PD_SUBSYSTEM, int event, int timeoutms);
int pd_notify_event(int board, PD_SUBSYSTEM, int event);

int pd_wait_for_event(int board, TSynchSS *sync, int timeoutms);
int pd_set_event(int board, TSynchSS *sync);


////////////////////////////////////////////////////////////////////////
//
// DRIVER INTERFACE SECTION
//
//

////////////////////////////////////////////////////////////////////////
//
//       NAME:  pd_bottom_half (dpc)
//
//   FUNCTION:  Handles the bottom half of PowerDAQ interrupts.
//
//  ARGUMENTS:  A board that needs attention.
//
//    RETURNS:  Nothing.
//
void pd_bottom_half(void *data)
{
   long board = (long)data;

   DPRINTK_T("i>bottom half got board %ld\n", board);

   _fw_spinlock    // set spin lock
   
   // check what happens and process events
   pd_process_events(board);
   
   // re-enable interrupts on this board
   pd_adapter_enable_interrupt(board, 1);
   
   _fw_spinunlock    // release spin lock
}

#if defined(_PD_RTL) || defined(_PD_RTLPRO)
void* pd_rt_bottom_half(void *data)
{
   int board = (int)data;

   DPRINTK_T("i>bottom half got board %d\n", board);

   // Wait for interrupts in a loop and notify the main thread
   // of the cause of the interrupt.
   while (1)
   {
      pthread_suspend_np(pthread_self());
      
      // call bottom-half
      pd_bottom_half(data);

      // Call user routine if registered
      // The call is protected by a mutex to syncronize it with the
      // calls to pd_register_user_isr() and pd_unregister_user_isr().
      pthread_mutex_lock(&pd_board_ext[board].user_isr_lock);

      if(pd_board_ext[board].user_isr != NULL)
      {
         pd_board_ext[board].user_isr(pd_board_ext[board].user_param);
      }

      pthread_mutex_unlock(&pd_board_ext[board].user_isr_lock);
   }
}
#elif defined(_PD_RTAI)
void pd_rt_bottom_half(int board)
{
   DPRINTK_T("i>bottom half got board %d\n", board);

   // Wait for interrupts in a loop and notify the main thread
   // of the cause of the interrupt.
   while (1)
   {
      rt_sem_wait (&rt_bh_sem[board]);
      
      // call bottom-half
      pd_bottom_half((void*)board);

      // Call user routine if registered
      // The call is protected by a mutex to syncronize it with the
      // calls to pd_register_user_isr() and pd_unregister_user_isr().
      rt_mutex_lock(&pd_board_ext[board].user_isr_lock);

      if(pd_board_ext[board].user_isr != NULL)
      {
         pd_board_ext[board].user_isr(pd_board_ext[board].user_param);
      }

      rt_mutex_unlock(&pd_board_ext[board].user_isr_lock);
   }
}
#endif



////////////////////////////////////////////////////////////////////////
//
//
//       NAME:  pd_isr_serve_board
//
//   FUNCTION:  Handles the PowerDAQ interrupts.  Called directly by the
//              main ISR for each board that uses the interrupt that
//              triggered.  Raises SIGIO to notify the userspace process
//              of asynchronous events.  Copies the first 512 samples out
//              of the sample FIFO to the user buffer.  Schedules the
//              bottom half.
//
//  ARGUMENTS:  A board that might need attention.
//
//    RETURNS:  1 if the board needed service and we provided it, and 0 if
//              the board did not need service.
//
int pd_isr_serve_board(int board)
{
   TEvents Events;

#ifdef MEASURE_LATENCY
   asm volatile
   (
       "rdtsc" : "=a" (pd_board[board].tscLow), "=d" (pd_board[board].tscHigh)
   );
#endif

   // check if the interrupt came from our adapter
   if ( !pd_dsp_int_status(board) )
   {
      // this board does not have an interrupt pending
      DPRINTK_Q("ISR: bogus interrupt! board=%d isr#%ld\n", board, intcnt);
      return 0;
   }

   _fw_spinlock    // set spin lock
   // acknowledge the interrupt
   if (!pd_dsp_acknowledge_interrupt(board))
      DPRINTK_F("isr: board %d not responding\n", board);

   if (pd_board[board].bTestInt == 1)
   {
      DPRINTK_Q("i>: test interrupt\n");
      pd_board[board].bTestInt = 0;
      if (pd_board[board].fasync)
      {
         //void pd_kill_fasync(struct fasync_struct *fa, int sig);
         if (pd_board[board].fasync)
            pd_kill_fasync(pd_board[board].fasync, SIGIO);
         DPRINTK_N("i>: testing interrupts...\n");
      }
      // get out of here
      _fw_spinunlock    // release spin lock
      return 1;
   }

   if (pd_board[board].bUseHeavyIsr && !(pd_board[board].dwXFerMode != XFERMODE_BM))
   {
      // here we should immediately read 512 samples from the FIFO
      pd_adapter_get_board_status(board, &Events);

      memcpy(&pd_board[board].FwEventsStatus, &Events, sizeof(TEvents));

      DPRINTK_T("i|AIB_FHF:%ld AIB_FHFSC:%ld\n",(Events.ADUIntr & AIB_FHF),(Events.AIOIntr & AIB_FHFSC));
      DPRINTK_T("i|AInIntr:0x%x\n", Events.AInIntr);

      if (pd_board[board].AinSS.bCheckHalfDone && 
          ((Events.ADUIntr & AIB_FHF) || (Events.AIOIntr & AIB_FHFSC)))
      { //AI:like my fix for WinNT
         DPRINTK_Q("i>isr-%ld: FHF flag is set, trying to flush FIFO\n", intcnt);

         if (!pd_ain_flush_fifo(board, 1))
         {
            DPRINTK_T("i>isr: No samples in the FIFO, sorry\n");
            _fw_spinunlock    // release spin lock
            return 1;
         } else
         {
            DPRINTK_T("i>isr: FIFO flashed with heavy interrupt\n");
         }
      }
   }

#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   // wake-up thread
   pthread_wakeup_np (rt_isr_thread[board]);

   // re-enable interrupt
   rtl_hard_enable_irq(pd_board[board].irq);
#elif defined(_PD_RTAI)
   rt_sem_signal (&rt_bh_sem[board]);

   // re-enable interrupt
   rt_enable_irq(pd_board[board].irq);
#else
   // schedule bottom half to run
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
   schedule_work(&pd_bh_task[board]);
#else
   pd_bh_task[board].data = (void*)board;
   queue_task(&pd_bh_task[board], &tq_immediate);
   mark_bh(IMMEDIATE_BH);
#endif
#endif
   DPRINTK_T("i>isr: DPC is scheduled!\n");

   _fw_spinunlock    // release spin lock

   return 1;
}

////////////////////////////////////////////////////////////////////////
//
//       NAME:  pd_isr
//
//   FUNCTION:  Handles the PowerDAQ interrupts.
//
//  ARGUMENTS:  The irq that triggered, and some other mumbo-jumbo.
//
//    RETURNS:  Nothing.
//
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
unsigned int pd_rt_isr(unsigned int irq, struct pt_regs *regs)
{
   int served;
   int board;

   DPRINTK_Q("i>got interrupt! irq=%d i#=%ld\n", irq, intcnt++);

   // see if any of our boards use this irq
   // if so, service all that do
   served = 0;
   for (board = 0; board < num_pd_boards; board ++)
   {
      if (pd_board[board].irq == irq)
      {
         if (pd_isr_serve_board(board))
         {
            served ++;
         }
      }
   }

   if(served == 0)
   {
      // This interrupt was not for us pass it on to the Linux kernel
      rtl_global_pend_irq(irq);
      rtl_hard_enable_irq(irq);
   }

   return 0;
}
#elif defined(_PD_RTAI)
void pd_rt_isr(int irq)
{
   int served;
   int board;
   
   DPRINTK_Q("i>got interrupt! irq=%d i#=%ld\n", irq, intcnt++);

   // see if any of our boards use this irq
   // if so, service all that do
   served = 0;
   for (board = 0; board < num_pd_boards; board ++)
   {
      if (pd_board[board].irq == irq)
      {
         if (pd_isr_serve_board(board))
         {
            served ++;
         }
      }
   }
}
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
irqreturn_t pd_isr(int trigger_irq, void *dev_id, struct pt_regs * regs)
#else
void pd_isr(int trigger_irq, void *dev_id, struct pt_regs * regs)
#endif
{
   int served;
   int board;

   DPRINTK_Q("i>got interrupt! irq=%d i#=%ld\n", trigger_irq, intcnt++);

   // see if any of our boards use this irq
   // if so, service all that do
   served = 0;
   for (board = 0; board < num_pd_boards; board ++)
   {
      if (&pd_board[board] == dev_id)
      {
         if (pd_isr_serve_board(board))
         {
            served ++;
         }
      }
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
   return ((served != 0) ? IRQ_HANDLED : IRQ_NONE);
#endif
}
#endif


//////////////////////////////////////////////////////////////////////////
//
//       name:  deal_with_device
//
//   function:  Checks if the passed-in device is one we know how to deal
//              with, and if so deals with it.
//
//  arguments:  The PCI device to deal with.  It is expected to have
//              Motorola's Vendor ID and the DSP56301's Device ID.
//
//    returns:  Nothing.
//
//
static void __init deal_with_device(struct pci_dev *dev) 
{
   u16 sub_vendor_id;
   u8  u8val;
   u16 u16val;
   u16 sub_device_id;
   int ret, t;
   char *model_name;


   if (dev == NULL)
   {
      DPRINTK_F("ERROR!  null PCI device passed to deal_with_device! skipping...\n");
      goto fail0;
   }

   if (dev->hdr_type != PCI_HEADER_TYPE_NORMAL)
   {
      DPRINTK_F("unexpected PCI header, skipping\n");
      return;
   }

   ret = pci_read_config_word(dev, PCI_SUBSYSTEM_VENDOR_ID, &sub_vendor_id);
   if (sub_vendor_id != UEI_SUBVENID)
   {
      DPRINTK_T("not a UEI board, skipping\n");
      return;
   }

   pci_read_config_word(dev, PCI_SUBSYSTEM_ID, &sub_device_id);
   DPRINTK_N("SubSystemID of board %d is 0x%x\n", num_pd_boards, sub_device_id);
   if ( (sub_device_id < PD_SUBSYSTEMID_FIRST) || (sub_device_id > PD_SUBSYSTEMID_LAST) )
   {
      DPRINTK_T("subdeviceid=%08X: this model of PowerDAQ board is not supported, skipping\n", sub_device_id);
      return;
   }

   DPRINTK_N("found a PowerDAQ board (bus=%d device=%d func=%d)\n", dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
   DPRINTK_N("Subsystem Id is 0x%x\n", sub_device_id);

   // initialize data structures
   intcnt = 0;

   memset(&pd_board[num_pd_boards], 0, sizeof(pd_board_t));

   pd_board[num_pd_boards].dev = dev;
   pd_board[num_pd_boards].caps_idx = sub_device_id - PD_SUBSYSTEMID_FIRST;

   DPRINTK_N("\tirq: %d\n", dev->irq);

   pd_board[num_pd_boards].size = 65536;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
   pd_board[num_pd_boards].address = ioremap(PCI_BASE_ADDRESS_MEM_MASK & dev->base_address[0], pd_board[num_pd_boards].size);
   DPRINTK_N("\tbus address: 0x%08lX\n", (unsigned long)dev->base_address[0]);
#else
   pd_board[num_pd_boards].address = ioremap(PCI_BASE_ADDRESS_MEM_MASK & dev->resource[0].start, pd_board[num_pd_boards].size);
   DPRINTK_N("\tbus address: 0x%08lX\n", (unsigned long)dev->resource[0].start);
#endif
   DPRINTK_N("\tvirt address: 0x%08lX\n", (unsigned long)pd_board[num_pd_boards].address);

   // store PCI configuration
   pci_read_config_word(dev, PCI_VENDOR_ID, &u16val);
   pd_board[num_pd_boards].PCI_Config.VendorID = u16val;

   pci_read_config_word(dev, PCI_DEVICE_ID, &u16val);
   pd_board[num_pd_boards].PCI_Config.DeviceID = u16val;

   pci_read_config_word(dev, PCI_COMMAND, &u16val);
   pd_board[num_pd_boards].PCI_Config.Command = u16val;

   pci_read_config_word(dev, PCI_STATUS, &u16val);
   pd_board[num_pd_boards].PCI_Config.Status = u16val;

   pci_read_config_byte(dev, PCI_REVISION_ID, &u8val);
   pd_board[num_pd_boards].PCI_Config.RevisionID = u8val;

   pci_read_config_byte(dev, PCI_CACHE_LINE_SIZE, &u8val);
   pd_board[num_pd_boards].PCI_Config.CacheLineSize = u8val;

   pci_read_config_byte(dev, PCI_LATENCY_TIMER, &u8val);
   pd_board[num_pd_boards].PCI_Config.LatencyTimer = u8val;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
   pd_board[num_pd_boards].PCI_Config.BaseAddress0 = dev->base_address[0];
#else
   pd_board[num_pd_boards].PCI_Config.BaseAddress0 = dev->resource[0].start;
#endif
   pd_board[num_pd_boards].PCI_Config.SubsystemVendorID = sub_vendor_id;
   pd_board[num_pd_boards].PCI_Config.SubsystemID = sub_device_id;

   pd_board[num_pd_boards].PCI_Config.InterruptLine = dev->irq;

   pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &u8val);
   pd_board[num_pd_boards].PCI_Config.InterruptPin = u8val;

   pci_read_config_byte(dev, PCI_MIN_GNT, &u8val);
   pd_board[num_pd_boards].PCI_Config.MinimumGrant = u8val;

   pci_read_config_byte(dev, PCI_MAX_LAT, &u8val);
   pd_board[num_pd_boards].PCI_Config.MaximumLatency = u8val;

   // check FW state and download it if necessarily
   if (!pd_dsp_startup (num_pd_boards)) goto fail1;

   pd_board[num_pd_boards].irq = dev->irq;
   pd_board[num_pd_boards].open = 0;

   // get hold on interrupt line
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   if (rqstirq)
   {
      // Grab the interrupt
      if(rtl_request_irq(pd_board[num_pd_boards].irq, pd_rt_isr)!= 0)
      {
         DPRINTK_F("couldnt allocate ISR, skipping card\n");
         goto fail1;
      }
   
      rtl_hard_enable_irq(pd_board[num_pd_boards].irq);
   }
   else
   {
      DPRINTK("rqstirq=0 interrupts disabled\n");
   }

   // initializes user ISR
   pd_board_ext[num_pd_boards].user_isr = NULL;
   pd_board_ext[num_pd_boards].user_param = NULL;
   // Initializes the mutex that protects access to the condition
   if(pthread_mutex_init(&pd_board_ext[num_pd_boards].user_isr_lock, NULL) != 0)
   {
      DPRINTK("deal_with_device: can't create mutex\n");
   }
#elif defined(_PD_RTAI)
   if (rqstirq)
   {
      // Grab the interrupt
      if(rt_request_global_irq(pd_board[num_pd_boards].irq, (void(*)(void))pd_rt_isr)!= 0)
      {
         DPRINTK_F("couldnt allocate ISR, skipping card\n");
         goto fail1;
      }
   
      rt_startup_irq(pd_board[num_pd_boards].irq);
      rt_enable_irq(pd_board[num_pd_boards].irq);
   }
   else
   {
      DPRINTK("rqstirq=0 interrupts disabled\n");
   }

   // initializes user ISR
   pd_board_ext[num_pd_boards].user_isr = NULL;
   pd_board_ext[num_pd_boards].user_param = NULL;
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board_ext[num_pd_boards].user_isr_lock);
#else
   if (rqstirq)
   {
      if (request_irq(pd_board[num_pd_boards].irq, pd_isr,
                      SA_SHIRQ | SA_INTERRUPT, "PowerDAQ",
                      (void *)&pd_board[num_pd_boards]))
      {
         DPRINTK_F("couldnt allocate ISR, skipping card\n");
         goto fail1;
      }
   } 
   else
   {
      DPRINTK("rqstirq=0 interrupts disabled\n");
   }
#endif

   pd_init_pd_board(num_pd_boards);

   // allocates memory for the synchronization data structure for each subsystem
   pd_board[num_pd_boards].AinSS.synch = (TSynchSS *) kmalloc(sizeof(TSynchSS), GFP_KERNEL);
   if (pd_board[num_pd_boards].AinSS.synch == NULL)
   {
      DPRINTK_T("Could not allocate memory for the Ain synch object.\n");
      goto fail1;
   }

   memset(pd_board[num_pd_boards].AinSS.synch, 0, sizeof(TSynchSS));

   pd_board[num_pd_boards].AoutSS.synch = (TSynchSS *) kmalloc(sizeof(TSynchSS), GFP_KERNEL);
   if (pd_board[num_pd_boards].AoutSS.synch == NULL)
   {
      DPRINTK_T("Could not allocate memory for the Aout synch object.\n");
      goto fail1;
   }

   memset(pd_board[num_pd_boards].AoutSS.synch, 0, sizeof(TSynchSS));

   pd_board[num_pd_boards].DinSS.synch = (TSynchSS *) kmalloc(sizeof(TSynchSS), GFP_KERNEL);
   if (pd_board[num_pd_boards].DinSS.synch == NULL)
   {
      DPRINTK_T("Could not allocate memory for the Dio synch object.\n");
      goto fail1;
   }

   memset(pd_board[num_pd_boards].DinSS.synch, 0, sizeof(TSynchSS));

   pd_board[num_pd_boards].DoutSS.synch = (TSynchSS *) kmalloc(sizeof(TSynchSS), GFP_KERNEL);
   if (pd_board[num_pd_boards].DoutSS.synch == NULL)
   {
      DPRINTK_T("Could not allocate memory for the Dio synch object.\n");
      goto fail1;
   }

   memset(pd_board[num_pd_boards].DoutSS.synch, 0, sizeof(TSynchSS));

   pd_board[num_pd_boards].UctSS.synch = (TSynchSS *) kmalloc(sizeof(TSynchSS), GFP_KERNEL);
   if (pd_board[num_pd_boards].UctSS.synch == NULL)
   {
      DPRINTK_T("Could not allocate memory for the Uct synch object.\n");
      goto fail1;
   }

   memset(pd_board[num_pd_boards].UctSS.synch, 0, sizeof(TSynchSS));


#if defined (_PD_RTL) || defined(_PD_RTLPRO)
   // Initializes the conditional variable used by the subsystem to notify
   // events
   ret = pthread_cond_init(&pd_board[num_pd_boards].AinSS.synch->event, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create conditional variable\n", ret);
      goto fail1;
   }

   // Initializes the mutex that protects access to the condition
   ret = pthread_mutex_init(&pd_board[num_pd_boards].AinSS.synch->event_lock, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create mutex\n", ret);
      goto fail1;
   }

   // Initializes the conditional variable used by the subsystem to notify
   // events
   ret = pthread_cond_init(&pd_board[num_pd_boards].AoutSS.synch->event, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create conditional variable\n", ret);
      goto fail1;
   }

   // Initializes the mutex that protects access to the condition
   ret = pthread_mutex_init(&pd_board[num_pd_boards].AoutSS.synch->event_lock, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create mutex\n", ret);
      goto fail1;
   }

   // Initializes the conditional variable used by the subsystem to notify
   // events
   ret = pthread_cond_init(&pd_board[num_pd_boards].DinSS.synch->event, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create conditional variable\n", ret);
      goto fail1;
   }

   // Initializes the mutex that protects access to the condition
   ret = pthread_mutex_init(&pd_board[num_pd_boards].DinSS.synch->event_lock, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create mutex\n", ret);
      goto fail1;
   }

   // Initializes the conditional variable used by the subsystem to notify
   // events
   ret = pthread_cond_init(&pd_board[num_pd_boards].DoutSS.synch->event, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create conditional variable\n", ret);
      goto fail1;
   }

   // Initializes the mutex that protects access to the condition
   ret = pthread_mutex_init(&pd_board[num_pd_boards].DoutSS.synch->event_lock, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create mutex\n", ret);
      goto fail1;
   }

   // Initializes the conditional variable used by the subsystem to notify
   // events
   ret = pthread_cond_init(&pd_board[num_pd_boards].UctSS.synch->event, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create conditional variable\n", ret);
      goto fail1;
   }

   // Initializes the mutex that protects access to the condition
   ret = pthread_mutex_init(&pd_board[num_pd_boards].UctSS.synch->event_lock, NULL);
   if (ret != 0)
   {
      DPRINTK("pwrdaq RT: Error %d, can't create mutex\n", ret);
      goto fail1;
   }
#elif defined(_PD_RTAI)
   // Initializes the conditional variable used by the subsystem to notify
   // events
   rt_cond_init(&pd_board[num_pd_boards].AinSS.synch->event);
   
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board[num_pd_boards].AinSS.synch->event_lock);
   
   // Initializes the conditional variable used by the subsystem to notify
   // events
   rt_cond_init(&pd_board[num_pd_boards].AoutSS.synch->event);
   
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board[num_pd_boards].AoutSS.synch->event_lock);
   
   // Initializes the conditional variable used by the subsystem to notify
   // events
   rt_cond_init(&pd_board[num_pd_boards].DinSS.synch->event);
   
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board[num_pd_boards].DinSS.synch->event_lock);
   
   // Initializes the conditional variable used by the subsystem to notify
   // events
   rt_cond_init(&pd_board[num_pd_boards].DoutSS.synch->event);
   
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board[num_pd_boards].DoutSS.synch->event_lock);
   
   // Initializes the conditional variable used by the subsystem to notify
   // events
   rt_cond_init(&pd_board[num_pd_boards].UctSS.synch->event);
      
   // Initializes the mutex that protects access to the condition
   rt_mutex_init(&pd_board[num_pd_boards].UctSS.synch->event_lock);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
   pd_board[num_pd_boards].AinSS.synch->wait_q = NULL;
   pd_board[num_pd_boards].AinSS.synch->sem = MUTEX;
   pd_board[num_pd_boards].AoutSS.synch->wait_q = NULL;
   pd_board[num_pd_boards].AoutSS.synch->sem = MUTEX;
   pd_board[num_pd_boards].DinSS.synch->wait_q = NULL;
   pd_board[num_pd_boards].DinSS.synch->sem = MUTEX;
   pd_board[num_pd_boards].DoutSS.synch->wait_q = NULL;
   pd_board[num_pd_boards].DoutSS.synch->sem = MUTEX;
   pd_board[num_pd_boards].UctSS.synch->wait_q = NULL;
   pd_board[num_pd_boards].UctSS.synch->sem = MUTEX;
#else   
   init_waitqueue_head(&pd_board[num_pd_boards].AinSS.synch->wait_q);
   init_waitqueue_head(&pd_board[num_pd_boards].AoutSS.synch->wait_q);
   init_waitqueue_head(&pd_board[num_pd_boards].DinSS.synch->wait_q);
   init_waitqueue_head(&pd_board[num_pd_boards].DoutSS.synch->wait_q);
   init_waitqueue_head(&pd_board[num_pd_boards].UctSS.synch->wait_q);
#endif
#endif // _PD_RTL

   pd_board[num_pd_boards].AinSS.synch->subsystem = AnalogIn;
   pd_board[num_pd_boards].AoutSS.synch->subsystem = AnalogOut;
   pd_board[num_pd_boards].DinSS.synch->subsystem = DigitalIn;
   pd_board[num_pd_boards].DoutSS.synch->subsystem = DigitalOut;
   pd_board[num_pd_boards].UctSS.synch->subsystem = CounterTimer;

   // Set the Xferm mode to whatever the user passed as a parameter
   pd_board[num_pd_boards].dwXFerMode = xferMode;
   DPRINTK_N("Xfer mode is %d\n", xferMode);

   // initialize calibration values
   pd_init_calibration(num_pd_boards);

   {
      u32 XBMPageSz, XAOPageSz;
      AllocContigMem Mem;
      
      // Verify that we do have enough memory for our games
      // load ADC FIFO transfer size
      /*if (( pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize >= 1 ) &&
          ( pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize <= 0x40 ))
      {
          pd_board[num_pd_boards].AinSS.FifoValues = pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize << 10; // kbytes->bytes
          
          // Limit the size of transfer to 0x2000 samples if the FIFO is bigger than that
          // this avoid holding the bus for too long and improve performances with
          // multiple boards.
          if(pd_board[num_pd_boards].AinSS.FifoValues <= PD_AIN_FIFO_VALUES)
             pd_board[num_pd_boards].AinSS.XferBufValues = pd_board[num_pd_boards].AinSS.FifoValues;
          else
             pd_board[num_pd_boards].AinSS.XferBufValues = PD_AIN_FIFO_VALUES;
      } 
      else 
      {
          pd_board[num_pd_boards].AinSS.FifoValues = PD_AIN_FIFO_VALUES;  // minimum size for 1kS FIFO
          pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize = 1;
      }*/
      // Verify that we do have enough memory for our games
      // load ADC FIFO transfer size
      if (( pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize >= 1 ) &&
          ( pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize <= 0x40 ))
      {
          pd_board[num_pd_boards].AinSS.FifoValues = pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize << 10; // kbytes->bytes
          if (pd_board[num_pd_boards].AinSS.FifoValues <= PD_AIN_MAX_FIFO_VALUES)
              pd_board[num_pd_boards].AinSS.XferBufValues = pd_board[num_pd_boards].AinSS.FifoValues;
          else
              pd_board[num_pd_boards].AinSS.XferBufValues = PD_AIN_MAX_FIFO_VALUES;    // maximum limited by transfer buffer
      }
      else
      {
          pd_board[num_pd_boards].AinSS.FifoValues = PD_AIN_FIFO_VALUES;  // minimum size for 1kS FIFO
          pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize = 1;
      }


      // calculate how many times we have to make 512 samples transfer. The idea is that we can make 1/2 ADC FIFO transfers
      pd_board[num_pd_boards].AinSS.FifoXFerCycles = (pd_board[num_pd_boards].AinSS.FifoValues / PD_AIN_FIFO_VALUES);

      // Set up bus-mastering pages parameters
      pd_board[num_pd_boards].AinSS.DoGetSamples = TRUE;

      // calculate this in pages: minimum - 1 page, maximum - x pages
      // please notice that XFer requres only one page where BM requires two
      XBMPageSz = 4;
      switch (pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize) 
      {
      case 1:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 1;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 8;
         break;
      case 2:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 2;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 8;
         break;
      case 4:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 4;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 8;
         break;
      case 8:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 8;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 8;
         break;
      case 16:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 16;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 16;
         XBMPageSz = 8;
         break;
      case 32:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 16;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 16;
         XBMPageSz = 8;
         break;
      case 64:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 16;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 32;
         XBMPageSz = 16;
         break;
      default:
         pd_board[num_pd_boards].AinSS.BmFHFXFers = 1;
         pd_board[num_pd_boards].AinSS.BmPageXFers = 8;
         break;
      }

      // Initialize memory chunks 0 and 1 for bus-mastering operations
      Mem.idx = AI_PAGE0;
      Mem.size = XBMPageSz;
      ret = pd_alloc_contig_memory(num_pd_boards, &Mem);
      if (ret != 0) 
      {
          pd_board[num_pd_boards].dwXFerMode = XFERMODE_NOAIMEM;
          pd_board[num_pd_boards].AinSS.XferBufValues = 0; 
          pd_board[num_pd_boards].AinSS.FifoValues = 0;
          pd_board[num_pd_boards].AinSS.pXferBuf = NULL;

          DPRINTK_F("pd_init ain error: couldn't allocate AIn buffer size 0x%x\n", XBMPageSz);

          goto fail1;
      } 
      else 
      {
          Mem.idx = AI_PAGE1;
          Mem.size = XBMPageSz;
          ret = pd_alloc_contig_memory(num_pd_boards, &Mem);
          if (ret != 0) 
          {
              pd_board[num_pd_boards].dwXFerMode = XFERMODE_NOAIMEM;
              pd_board[num_pd_boards].AinSS.XferBufValues = 0;
              pd_board[num_pd_boards].AinSS.FifoValues = 0;
              pd_board[num_pd_boards].AinSS.pXferBuf = NULL;
              goto fail1;
          }
          pd_board[num_pd_boards].AinSS.pXferBuf = pd_board[num_pd_boards].pSysBMB[AI_PAGE1];  // buffer address for XFer and BM
      }

      // load DAC FIFO transfer size
      if (pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize > 0x80) // incorrect DA88C FIFO size
          pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize = 2;

      if (( pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize >= 2 ) &&
          ( pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize <= 0x40 ))
      {
          pd_board[num_pd_boards].AoutSS.FifoValues = pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize << 10; // kbytes->bytes
          if (pd_board[num_pd_boards].AoutSS.FifoValues <= PD_AOUT_MAX_FIFO_VALUES)
              pd_board[num_pd_boards].AoutSS.XferBufValues = pd_board[num_pd_boards].AoutSS.FifoValues;
          else
              pd_board[num_pd_boards].AoutSS.XferBufValues = ANALOG_XFERBUF_VALUES;    // maximum limited by transfer buffer
      } 
      else 
         {
          pd_board[num_pd_boards].AoutSS.FifoValues = PD_AOUT_MAX_FIFO_VALUES;  // minimum size for 1kS FIFO
      }

      pd_board[num_pd_boards].AoutSS.TranSize = pd_board[num_pd_boards].AoutSS.FifoValues;

      // calculate how many pages need to be allocated
      XAOPageSz = (pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize << 12)/PAGE_SIZE;
      if ((pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize << 12)%PAGE_SIZE) 
         XAOPageSz++;

      Mem.idx = AO_PAGE0;
      Mem.size = XAOPageSz;
      ret = pd_alloc_contig_memory(num_pd_boards, &Mem);
      if (ret != 0) 
      {
         pd_board[num_pd_boards].AoutSS.TranSize = 0;
         pd_board[num_pd_boards].AoutSS.FifoValues = 0;
         pd_board[num_pd_boards].AoutSS.pXferBuf = NULL;
         pd_board[num_pd_boards].dwXFerMode = XFERMODE_NOAOMEM;

         DPRINTK_F("pd_init AO/AO32 error: couldn't allocate AOut buffer size 0x%x\n", XAOPageSz);
      }
      pd_board[num_pd_boards].AoutSS.pXferBuf = pd_board[num_pd_boards].pSysBMB[AO_PAGE0];
   }

   DPRINTK_N("AI: XferBufValues = %d, BlkXferValues = %d, FifoValues = %d, FifoXFerCycles = %d\n", 
                    pd_board[num_pd_boards].AinSS.XferBufValues,
                    pd_board[num_pd_boards].AinSS.BlkXferValues, 
                    pd_board[num_pd_boards].AinSS.FifoValues, 
                    pd_board[num_pd_boards].AinSS.FifoXFerCycles);


   DPRINTK_N("AI0:0x%lx AI1:0x%lx AO0:0x%lx\n", (unsigned long)pd_board[num_pd_boards].pSysBMB[AI_PAGE0], 
                                           (unsigned long)pd_board[num_pd_boards].pSysBMB[AI_PAGE1], 
                                           (unsigned long)pd_board[num_pd_boards].pSysBMB[AO_PAGE0]);

   DPRINTK_N("pd_board[%d] has been initialized\n", num_pd_boards);
   
   if(PD_IS_PDXI(pd_board[num_pd_boards].PCI_Config.SubsystemID))
      model_name = pdxi_models[pd_board[num_pd_boards].PCI_Config.SubsystemID - 0x211].modelname;
   else
      model_name = pd_models[pd_board[num_pd_boards].PCI_Config.SubsystemID - 0x101].modelname;

   printk("Board %d:\n", num_pd_boards);
   printk("\tName: %s\n", model_name);
   printk("\tSerial Number: %8s\n", pd_board[num_pd_boards].Eeprom.u.Header.SerialNumber);
   printk("\tInput FIFO size: %d samples\n", pd_board[num_pd_boards].Eeprom.u.Header.ADCFifoSize*1024);
   printk("\tInput channel list FIFO size: %d entries\n", pd_board[num_pd_boards].Eeprom.u.Header.CLFifoSize*256);
   printk("\tIOutut FIFO size: %d samples\n", pd_board[num_pd_boards].Eeprom.u.Header.DACFifoSize*1024);
   printk("\tManufacture date: %s\n", pd_board[num_pd_boards].Eeprom.u.Header.ManufactureDate);
   printk("\tCalibration date: %s\n", pd_board[num_pd_boards].Eeprom.u.Header.CalibrationDate);
   printk("\tBase address: 0x%lx\n", pd_board[num_pd_boards].PCI_Config.BaseAddress0);
   printk("\tIRQ line: 0x%x\n", pd_board[num_pd_boards].PCI_Config.InterruptLine); 
   printk("\tDSP Rev: v%d\n", (pd_board[num_pd_boards].fwTimestamp[0]&0x0F0000)>>16);  
   t =  pd_board[num_pd_boards].fwTimestamp[0]>>20;
   printk("\tFirmware type: %s, rev: %d.%d/%x\n",
          (t < 1)?"Generic": ((t<2)?"AO": ((t<3)?"DIO":((t<4)?"LMF":((t<5)?"MFx":"Unknown")))),
          (pd_board[num_pd_boards].fwTimestamp[0]&0xF00)>>8,
          pd_board[num_pd_boards].fwTimestamp[0]&0xFF,
          pd_board[num_pd_boards].fwTimestamp[1]); 

   num_pd_boards ++;
   return;

   fail1:  iounmap(pd_board[num_pd_boards].address);
   fail0:  return;
}

//////////////////////////////////////////////////////////////////////////
//
//       NAME: pd_open()
//
//   FUNCTION: Open a PowerDAQ device.
//
//  ARGUMENTS: Hmmm..
//
//    RETURNS: 0 if all went well, and some error code if there was a problem.
//
int pd_open(struct inode *inode, struct file *file)
{
   int real_minor, board, board_minor;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   DPRINTK_I("trying to open board %d, %s (minor %u)\n",
             board, pd_devices_by_minor[board_minor], real_minor);

   switch (board_minor)
   {
   case PD_MINOR_AIN:
      if (!pd_board[board].AinSS.bInUse)
      {
         pd_board[board].AinSS.bInUse = TRUE;
         pd_board[board].AinSS.iPID = current->pid;
         pd_board[board].fd[AnalogIn] = 0;
      } else
      {
         DPRINTK_I("pd_open error: AinSS is in use\n");
         return -EBUSY;
      }
      break;

   case PD_MINOR_AOUT:
      if (!pd_board[board].AoutSS.bInUse)
      {
         pd_board[board].AoutSS.dwAoutValue = 0x800800;
         pd_board[board].AoutSS.bInUse = TRUE;
         pd_board[board].AoutSS.iPID = current->pid;
         pd_board[board].fd[AnalogOut] = 0;
      } else
      {
         DPRINTK_I("pd_open error: AoutSS is in use\n");
         return -EBUSY;
      }
      break;

   case PD_MINOR_DIN:
      if (!pd_board[board].DinSS.bInUse)
      {
         pd_board[board].DinSS.bInUse = TRUE;
         pd_board[board].DinSS.iPID = current->pid;
         pd_board[board].fd[DigitalIn] = 0;
      } else
      {
         DPRINTK_I("pd_open error: DinSS is in use\n");
         return -EBUSY;
      }
      break;

   case PD_MINOR_DOUT:
      if (!pd_board[board].DoutSS.bInUse)
      {
         pd_board[board].DoutSS.bInUse = TRUE;
         pd_board[board].DoutSS.iPID = current->pid;
         pd_board[board].fd[DigitalOut] = 0;
      } else
      {
         DPRINTK_I("pd_open error: DoutSS is in use\n");
         return -EBUSY;
      }
      break;

   case PD_MINOR_UCT:
      if (!pd_board[board].UctSS.bInUse)
      {
         pd_board[board].UctSS.bInUse = TRUE;
         pd_board[board].UctSS.iPID = current->pid;
         pd_board[board].fd[CounterTimer] = 0;
      } else
      {
         DPRINTK_I("pd_open error: UctSS is in use\n");
         return -EBUSY;
      }
      break;

   case PD_MINOR_DSPCT:
      if (!pd_board[board].UctSS.bInUse)
      {
         pd_board[board].UctSS.bInUse = TRUE;
         pd_board[board].UctSS.iPID = current->pid;
         pd_board[board].fd[DSPCounter] = 0;
      } else
      {
         DPRINTK_I("pd_open error: UctSS is in use\n");
         return -EBUSY;
      }
      break;

   }

   // test interrupts
   /*if (!pd_board[board].open)
   {
      retf = pd_ain_reset(board);
      if (!retf)
         DPRINTK_I("pd_adapter_disable_interrupt failed\n");
   } */

   pd_board[board].open++;
   DPRINTK_I("pd_open: open count = %d\n", pd_board[board].open);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
   MOD_INC_USE_COUNT;
#endif

   return 0;
}


//////////////////////////////////////////////////////////////////////////
//
//       NAME:  pd_release()
//
//   FUNCTION:  Frees up a PowerDAQ device.
//
//  ARGUMENTS:  Hmmm..
//
//    RETURNS:  None.
//
int pd_release(
              struct inode *inode,
              struct file *file
              )
{
   PTBuf_Info pDaqBuf = NULL;
   int real_minor, board, board_minor, subsystem;
   int pd_fasync(int fd, struct file *file, int mode);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;
   subsystem = BoardLevel;

   DPRINTK_I("pd_release: board %d, (minor %u)\n",
             board, real_minor);

   // if release is called, library shall already inform driver about it
   // using IOCTL_PWRDAQ_CLOSESUBSYSTEM, so clearing of fd seems to be safe
   switch (board_minor)
   {
   case PD_MINOR_AIN:
      if (pd_board[board].AinSS.bInUse)
      {
         pd_board[board].AinSS.bInUse = FALSE;
         pd_board[board].AinSS.iPID = 0;
         pDaqBuf = &pd_board[board].AinSS.BufInfo;
         pd_board[board].fd[AnalogIn] = 0;
         subsystem = AnalogIn;
      }
      break;

   case PD_MINOR_AOUT:
      if (pd_board[board].AoutSS.bInUse)
      {
         pd_board[board].AoutSS.bInUse = FALSE;
         pd_board[board].AoutSS.iPID = 0;
         pDaqBuf = &pd_board[board].AoutSS.BufInfo;
         pd_board[board].fd[AnalogOut] = 0;
         subsystem = AnalogOut;
      }
      break;

   case PD_MINOR_DIN:
      if (pd_board[board].DinSS.bInUse)
      {
         pd_board[board].DinSS.bInUse = FALSE;
         pd_board[board].DinSS.iPID = 0;
         pd_board[board].fd[DigitalIn] = 0;
         subsystem = DigitalIn;
      }
      break;

   case PD_MINOR_DOUT:
      if (pd_board[board].DoutSS.bInUse)
      {
         pd_board[board].DoutSS.bInUse = FALSE;
         pd_board[board].DoutSS.iPID = 0;
         pd_board[board].fd[DigitalOut] = 0;
         subsystem = DigitalOut;
      }
      break;

   case PD_MINOR_UCT:
      if (pd_board[board].UctSS.bInUse)
      {
         pd_board[board].UctSS.bInUse = FALSE;
         pd_board[board].UctSS.iPID = 0;
         pd_board[board].fd[CounterTimer] = 0;
         subsystem = CounterTimer;
      }
      break;

   case PD_MINOR_DSPCT:
      if (pd_board[board].UctSS.bInUse)
      {
         pd_board[board].UctSS.bInUse = FALSE;
         pd_board[board].UctSS.iPID = 0;
         pd_board[board].fd[DSPCounter] = 0;
         subsystem = DSPCounter;
      }
      break;

   }

   // possibly our app crashes. We need to deallocate buffers...
   if (pDaqBuf && pDaqBuf->databuf) // AIn/AOut and buffer is allocated...
      pd_unregister_daq_buffer(board, subsystem);

   // release any asynchronous readers, fd not used since fasync stuct being released
   pd_fasync( -1, file, 0);  //FIXME: shouldn't I release it for each subsystem???

   // if last board closed, free the IRQ
   pd_board[board].open--;
   DPRINTK_I("pd_close: open count = %d\n", pd_board[board].open);
   if (!pd_board[board].open)
   {
      pd_board[board].AinSS.bInUse = FALSE;
      pd_board[board].AoutSS.bInUse = FALSE;
      pd_board[board].DinSS.bInUse = FALSE;
      pd_board[board].DoutSS.bInUse = FALSE;
      pd_board[board].UctSS.bInUse = FALSE;
   }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
   MOD_DEC_USE_COUNT;
#endif

   return 0;
}

////////////////////////////////////////////////////////////////////////
//
// FILE OPERATIONS SECTION
//
////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////
//
//       Name:  pd_read()
//
//   Function:  Reads data from the board.
//
//  Arguments:  The file being read from, the userspace buffer to read into,
//              the number of bytes to read, and the offset (what's this?)...
//
//    Returns:  Number of bytes actually read.
//
ssize_t pd_read(
               struct file *file,
               char *buffer,
               size_t count,
               loff_t *ppos )
{
   int real_minor, board, board_minor;
   int ret = -ENODEV;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   DPRINTK_I("trying to read from board %d, %s (minor %u)\n", board,
             pd_devices_by_minor[board_minor], real_minor);

   /*if (!CheckOwnerPID(board, board_minor, current->pid))
   {
      DPRINTK_F("Ownership check failed: PID:%i\n", current->pid);
      return -EBUSY;
   } */

   _fw_spinlock    // set spin lock

   switch (board_minor)
   {
   case PD_MINOR_AIN:
      if (!(pd_board[board].AinSS.SubsysState & ssActive))
      {
         ret = pd_ain_read_scan(board, buffer, count);
      }
      break;

   case PD_MINOR_AOUT:
      if (!(pd_board[board].AoutSS.SubsysState & ssActive))
      {
         DPRINTK_I("not implemented yet...\n");
      }
      ret = -ENODEV;
      break;

   case PD_MINOR_DIN:
      if (!(pd_board[board].DinSS.SubsysState & ssActive))
      {
         ret = pd_din_read(board, buffer, count);
      }
      break;

   case PD_MINOR_DOUT:
      if (!(pd_board[board].DoutSS.SubsysState & ssActive))
      {
         ret = pd_din_read(board, buffer, count);
      }
      break;

   case PD_MINOR_UCT:
      if (!(pd_board[board].UctSS.SubsysState & ssActive))
      {
         DPRINTK_I("not implemented yet...\n");
      }
      ret = -ENODEV;
      break;
   }

   _fw_spinunlock    // release spin lock

   return ret;
}


///////////////////////////////////////////////////////////////////////
//
//       Name:  pd_write()
//
//   Function:  Writes data to the board.
//
//  Arguments:  Hmmm...
//
//    Returns:  Number of bytes actually written.
//
ssize_t pd_write(
                struct file *file,
                const char *buffer,
                size_t count,
                loff_t *ppos )
{
   int real_minor, board, board_minor;
   int ret = -ENODEV;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   DPRINTK_I("trying to write to board %d, %s (minor %u)\n",
             board, pd_devices_by_minor[board_minor], real_minor);

   /*if (!CheckOwnerPID(board, board_minor, current->pid))
   {
      DPRINTK_F("Ownership check failed: PID:%i\n", current->pid);
      return -EBUSY;
   } */

   _fw_spinlock    // set spin lock

   switch (board_minor)
   {
   
   case PD_MINOR_AIN:
      if (!(pd_board[board].AinSS.SubsysState & ssActive))
      {
         DPRINTK_I("not implemented yet...\n");
      }
      ret = -ENODEV;
      break;

   case PD_MINOR_AOUT:
      if (!(pd_board[board].AoutSS.SubsysState & ssActive))
      {
         ret = pd_aout_write_scan(board, buffer, count);
      }
      break;

   case PD_MINOR_DIN:
      if (!(pd_board[board].DinSS.SubsysState & ssActive))
      {
         ret = pd_dout_write(board, buffer, count);
      }
      break;

   case PD_MINOR_DOUT:
      if (!(pd_board[board].DoutSS.SubsysState & ssActive))
      {
         ret = pd_dout_write(board, buffer, count);
      }
      break;

   case PD_MINOR_UCT:
      if (!(pd_board[board].UctSS.SubsysState & ssActive))
      {
         DPRINTK_I("not implemented yet...\n");
      }
      ret = -ENODEV;
      break;
   }

   _fw_spinunlock    // release spin lock

   return ret;
}

///////////////////////////////////////////////////////////////////////
//
// DISPATCH ROUTINE
//
///////////////////////////////////////////////////////////////////////
//
//       Name:  pd_ioctl / pd_rt_ioctl
//
//   Function:  Special functions for configuring the driver subsystems.
//
//  Arguments:  Minor doesn't matter ... really
//
//    Returns:  0 if everything went well, and "what errno should be" if
//              there's a problem.
//
//    Note:     Majority of FW functions have another return code notation.
//              FW returns 1 if SUCCESS and 0 if FAILURE. Don't forget
//              to translate them
//
//
#if defined (_PD_RTL) || defined(_PD_RTAI) || defined(_PD_RTLPRO)
int pd_ioctl(struct inode *inode, struct file *file,
             unsigned int command, unsigned long argument);

int pd_rt_ioctl(int board, unsigned int command, unsigned long argument)
{
   struct inode* pinode = NULL;
   struct file m_file;
   struct file* pfile = &m_file;
   int retcode;

   // fill out whatever needs to be filled out for pd_ioctl
   pfile->f_dentry->d_inode->i_rdev = board * PD_MINOR_RANGE;

   // call it!
   retcode = pd_ioctl(pinode, pfile, command, argument);
   return retcode;
}

#endif

int pd_ioctl(
            struct inode *inode,
            struct file *file,
            unsigned int command,
            unsigned long arg)
{
   int real_minor, board, board_minor, ret;
   PD_SUBSYSTEM ss;
   int retf = -ENODEV;
   int tout, i;
   unsigned long id;
   TCmd argcmd;

   DPRINTK_I("pd_ioctl: Process:%s PID:%d\n", current->comm, current->pid);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   if (board_minor == PD_MINOR_AIN)
      ss = AnalogIn;
   else if (board_minor == PD_MINOR_AOUT)
      ss = AnalogOut;
   else if (board_minor == PD_MINOR_DIN)
      ss = DigitalIn;
   else if (board_minor == PD_MINOR_DOUT)
      ss = DigitalOut;
   else if (board_minor == PD_MINOR_UCT)
      ss = CounterTimer;
   else if (board_minor == PD_MINOR_DSPCT)
      ss = DSPCounter;
   else if (board_minor == PD_MINOR_DRV)
      ss = BoardLevel;
   else
      ss = BoardLevel;

   DPRINTK_I("board %d, %s (minor %u) got ioctl 0x%X, argument %ld...\n",
             board, pd_devices_by_minor[board_minor], real_minor, command, arg);

   if (PD_IS_PDXI(pd_board[board].PCI_Config.SubsystemID))
      id = pd_board[board].PCI_Config.SubsystemID - 0x100;
   else
      id = pd_board[board].PCI_Config.SubsystemID;

   // check for board model before executing IOCTL
   if (PD_IS_DIO(id))
   {
      switch (command)
      {
      case IOCTL_PWRDAQ_DOWRITE:
      case IOCTL_PWRDAQ_DORESET:

      case IOCTL_PWRDAQ_UCTSETCFG:
      case IOCTL_PWRDAQ_UCTSTATUS:
      case IOCTL_PWRDAQ_UCTWRITE:
      case IOCTL_PWRDAQ_UCTREAD:
      case IOCTL_PWRDAQ_UCTSWGATE:

      //case IOCTL_PWRDAQ_DISETCFG:
      case IOCTL_PWRDAQ_DISTATUS:
      case IOCTL_PWRDAQ_DICLRDATA:
      case IOCTL_PWRDAQ_DIREAD:
      case IOCTL_PWRDAQ_DIRESET:

      case IOCTL_PWRDAQ_CALSETCFG:
      case IOCTL_PWRDAQ_CALDACWRITE:
         retf = -ENOSYS;
         return retf;
         break;
      }
   } 
   else if (PD_IS_AO(id))
   {
      switch (command)
      {
      case IOCTL_PWRDAQ_UCTSETCFG:
      case IOCTL_PWRDAQ_UCTSTATUS:
      case IOCTL_PWRDAQ_UCTWRITE:
      case IOCTL_PWRDAQ_UCTREAD:
      case IOCTL_PWRDAQ_UCTSWGATE:
      case IOCTL_PWRDAQ_UCTSWCLK:
      case IOCTL_PWRDAQ_UCTRESET:

      case IOCTL_PWRDAQ_AISETCFG:
      case IOCTL_PWRDAQ_AISETCVCLK:
      case IOCTL_PWRDAQ_AISETCLCLK:
      case IOCTL_PWRDAQ_AISETCHLIST:
      case IOCTL_PWRDAQ_AISETEVNT:
      case IOCTL_PWRDAQ_AISTATUS:
      case IOCTL_PWRDAQ_AICVEN:
      case IOCTL_PWRDAQ_AISTARTTRIG:
      case IOCTL_PWRDAQ_AISTOPTRIG:
      case IOCTL_PWRDAQ_AISWCLSTART:
      case IOCTL_PWRDAQ_AICLRESET:
      case IOCTL_PWRDAQ_AICLRDATA:
      case IOCTL_PWRDAQ_AIRESET:
      case IOCTL_PWRDAQ_AIGETVALUE:
      case IOCTL_PWRDAQ_AIGETSAMPLES:
      case IOCTL_PWRDAQ_AISETSSHGAIN:
      case IOCTL_PWRDAQ_AISETXFERSIZE:
      //case IOCTL_PWRDAQ_AIN_BLK_XFER:
         retf = -ENOSYS;
         return retf;
         break;
      }
   } 
   else if (PD_IS_LABMF(id))
   {
      switch (command)
      {
      case IOCTL_PWRDAQ_UCTSETCFG:
      case IOCTL_PWRDAQ_UCTSTATUS:
      case IOCTL_PWRDAQ_UCTWRITE:
      case IOCTL_PWRDAQ_UCTREAD:
      case IOCTL_PWRDAQ_UCTSWGATE:
      case IOCTL_PWRDAQ_UCTSWCLK:
      case IOCTL_PWRDAQ_UCTRESET:
         retf = -ENOSYS;
         return retf;
         break;
      }
   }

   if((void*)arg != NULL)
      pd_copy_from_user32((u32*)&argcmd, (u32*)arg, sizeof(TCmd));

   _fw_spinlock

   switch (command)
   {
   case  IOCTL_PWRDAQ_OPENSUBSYSTEM:
      if (pd_board[board].fd[argcmd.AcqSS.subSystem] == 0)
      {
         pd_board[board].fd[argcmd.AcqSS.subSystem] = argcmd.AcqSS.fileDescriptor;
         retf = 0;
         DPRINTK_I("IOCTL_PWRDAQ_OPENSUBSYSTEM with %d,%d\n",
                   argcmd.AcqSS.subSystem,argcmd.AcqSS.fileDescriptor);
      } 
      else
      {
         DPRINTK_I("IOCTL_PWRDAQ_OPENSUBSYSTEM: board %d susbsystem %s is busy\n", board, pd_devices_by_minor[board_minor]); 
         retf = -EBUSY;
      }
      break;

   case  IOCTL_PWRDAQ_CLOSESUBSYSTEM:
      if (pd_board[board].fd[argcmd.AcqSS.subSystem] == 
          argcmd.AcqSS.fileDescriptor)
      {
         pd_board[board].fd[argcmd.AcqSS.subSystem] = 0;
         retf = 0;
         DPRINTK_I("IOCTL_PWRDAQ_CLOSESUBSYSTEM with %d,%d\n",
                   argcmd.AcqSS.subSystem, argcmd.AcqSS.fileDescriptor);

      } 
      else
      {
         DPRINTK_I("IOCTL_PWRDAQ_CLOSESUBSYSTEM: board %d susbsystem %s file descriptor doesn'tmatch\n",board,pd_devices_by_minor[board_minor]); 
         retf = -EBUSY;
      }
      break;

   case  IOCTL_PWRDAQ_RELEASEALL:
      retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_PRIVATE_GETCFG:
      memcpy(&argcmd.PciConfig, (u8*)&pd_board[board].PCI_Config, sizeof(PD_PCI_CONFIG));
      retf = 0;
      break;

   case  IOCTL_PWRDAQ_PRIVATE_SLEEP_ON:
      tout = (DWORD)argcmd.dwParam[0];
      tout = (tout * HZ)/1000;
      retf = interruptible_sleep_on_timeout(&whead, tout); 
      break;

   case  IOCTL_PWRDAQ_PRIVATE_WAKE_UP:
      retf = 0;
      wake_up_interruptible(&whead);
      break;

   case  IOCTL_PWRDAQ_PRIVATE_SET_EVENT:  // put to sleep...
      // store timeout for subsystem
      retf = pd_sleep_on_event(board, ss,
                               argcmd.dwParam[0],  //event
                               argcmd.dwParam[1]); //timeout
      break;

   case IOCTL_PWRDAQ_PRIVATE_CLR_EVENT:
      // , // Event
      // for now - wake up from event
      retf = pd_notify_event(board, ss, argcmd.dwParam[0]);

      break;

   case  IOCTL_PWRDAQ_GET_NUMBER_ADAPTER:
      argcmd.dwParam[0] = num_pd_boards;
      retf = 0;
      break;

   case  IOCTL_PWRDAQ_REGISTER_BUFFER:
      // Release spinlock, the code in pd_register_daq_buffer
      // is not safe to run with spinlock held
      _fw_spinunlock

      retf = argcmd.dwParam[0] = pd_register_daq_buffer(board,
                                    argcmd.dwParam[4], // subsystem
                                    argcmd.dwParam[0], // dwScanSize
                                    argcmd.dwParam[1], // dwScansFrm
                                    argcmd.dwParam[2], // dwFramesBfr
                                    NULL, // ptr to user buffer
                                    argcmd.dwParam[3]);

      // Reaquire the lock
      _fw_spinlock

      retf = (retf) ? retf : -EIO;
      break;

#if defined(_PD_RTL) || defined(_PD_RTAI) || defined(_PD_RTLPRO)
   case  IOCTLRT_PWRDAQ_GETKERNELBUFPTR:
      if (board_minor == PD_MINOR_AIN)   // Note: kernel-to-kernel space request
         argcmd.dwParam[0] = (u32)pd_board[board].AinSS.BufInfo.databuf;
      else if (board_minor == PD_MINOR_AOUT)
         argcmd.dwParam[0] = (u32)pd_board[board].AoutSS.BufInfo.databuf;
      else argcmd.dwParam[0] = -EINVAL;
      break;
#endif

   case  IOCTL_PWRDAQ_UNREGISTER_BUFFER:
      // Release spinlock, the code in pd_unregister_daq_buffer
      // is not safe to run with spinlock held
      _fw_spinunlock

      retf = argcmd.dwParam[1] =
             pd_unregister_daq_buffer(board, argcmd.dwParam[0]);

      // Reaquire the lock
      _fw_spinlock

      retf = (retf) ? 0 : -EIO;
      break;

   case IOCTL_PWRDAQ_GETKERNELBUFSIZE:
      retf = argcmd.dwParam[1] =
              pd_get_daq_buffer_size(board, argcmd.dwParam[0]);
      retf = (retf) ? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_SET_USER_EVENTS:
      if (PD_IS_DIO(id) && !(argcmd.dwParam[0] & EdgeDetect))
      {
         if (argcmd.dwParam[0] == DigitalIn)
            argcmd.dwParam[0] = AnalogIn;
         if (argcmd.dwParam[0] == DigitalOut)
            argcmd.dwParam[0] = AnalogOut;
         if (argcmd.dwParam[0] == DSPCounter)
            argcmd.dwParam[0] = AnalogIn;
      } else
      {
         argcmd.dwParam[0] = argcmd.dwParam[0] & 0xF;
      }

      retf = (pd_set_user_events(board,
                                 argcmd.dwParam[0],
                                 argcmd.dwParam[1]) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_CLEAR_USER_EVENTS:
      if (PD_IS_DIO(id) && !(argcmd.dwParam[0] & EdgeDetect))
      {
         if (argcmd.dwParam[0] == DigitalIn)
            argcmd.dwParam[0] = AnalogIn;
         if (argcmd.dwParam[0] == DigitalOut)
            argcmd.dwParam[0] = AnalogOut;
         if (argcmd.dwParam[0] == CounterTimer)
            argcmd.dwParam[0] = AnalogIn;
         if (argcmd.dwParam[0] == DSPCounter)
            argcmd.dwParam[0] = AnalogIn;
      } else
      {
         argcmd.dwParam[0] = argcmd.dwParam[0] & 0xF;
      }

      retf = (pd_clear_user_events(board,
                                   argcmd.dwParam[0],
                                   argcmd.dwParam[1])? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_GET_USER_EVENTS:
      if (PD_IS_DIO(id) && !(argcmd.dwParam[0] & EdgeDetect))
      {
         if (argcmd.dwParam[0] == DigitalIn)
            argcmd.dwParam[0] = AnalogIn;
         if (argcmd.dwParam[0] == DigitalOut)
            argcmd.dwParam[0] = AnalogOut;
         if (argcmd.dwParam[0] == DSPCounter)
            argcmd.dwParam[0] = AnalogIn;
      } else
      {
         argcmd.dwParam[0] = argcmd.dwParam[0] & 0xF;
      }

      retf = (pd_get_user_events(board,
                                 argcmd.dwParam[0],
                                 (u32*)&argcmd.dwParam[1]) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_IMMEDIATE_UPDATE:
      retf = (pd_immediate_update(board))? 0 : -EIO;
      break;


   case  IOCTL_PWRDAQ_SET_TIMED_UPDATE: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_AIN_ASYNC_INIT:
      retf = (pd_ain_async_init(board, &(argcmd.AinAsyncCfg)) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_AIN_ASYNC_TERM:
      retf = (pd_ain_async_term(board) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_AIN_ASYNC_START:
      if (pd_ain_async_start(board))
         retf = 0;
      else retf = -EIO;
      break;

   case  IOCTL_PWRDAQ_AIN_ASYNC_STOP:
      if (pd_ain_async_stop(board))
         retf = 0;
      else retf = -EIO;
      break;

   case  IOCTL_PWRDAQ_AO_ASYNC_INIT:
      retf = (pd_aout_async_init(board, &(argcmd.AinAsyncCfg)) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_AO_ASYNC_TERM:
      retf = (pd_aout_async_term(board) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_AO_ASYNC_START:
      if (pd_aout_async_start(board))
         retf = 0;
      else retf = -EIO;
      break;

   case  IOCTL_PWRDAQ_AO_ASYNC_STOP:
      if (pd_aout_async_stop(board))
         retf = 0;
      else retf = -EIO;
      break;

   case  IOCTL_PWRDAQ_GET_DAQBUF_STATUS: retf = -ENOSYS;
      //pd_ain_async_get_status(board); // not for now, OK?
      break;

   case  IOCTL_PWRDAQ_GET_DAQBUF_SCANS:
      if ((argcmd.ScanInfo.Subsystem == AnalogIn)||
          (argcmd.ScanInfo.Subsystem == DigitalIn)||
          (argcmd.ScanInfo.Subsystem == CounterTimer)||
          (argcmd.ScanInfo.Subsystem == DSPCounter))
         retf = pd_ain_get_scans( board, &argcmd.ScanInfo) ? 0 : -EIO;   // input buffer
      else if ((argcmd.ScanInfo.Subsystem == AnalogOut)||
               (argcmd.ScanInfo.Subsystem == DigitalOut))
         retf = pd_aout_get_scans( board, &argcmd.ScanInfo) ? 0 : -EIO; // output buffer

      break;

   case  IOCTL_PWRDAQ_CLEAR_DAQBUF: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_BRDRESET: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_BRDEEPROMREAD:
      if (argcmd.EepromAcc.MaxSize > PD_EEPROM_SIZE)
         argcmd.EepromAcc.MaxSize = PD_EEPROM_SIZE;

      memcpy(argcmd.EepromAcc.Buffer, pd_board[board].Eeprom.u.WordValues,
             argcmd.EepromAcc.MaxSize*sizeof(u16));
      
      argcmd.EepromAcc.WordsRead = argcmd.EepromAcc.MaxSize;
      retf = argcmd.EepromAcc.WordsRead;
      break;

   case  IOCTL_PWRDAQ_BRDEEPROMWRITE:
      retf = (pd_adapter_eeprom_write(board,
                                      argcmd.EepromAcc.MaxSize,
                                      argcmd.EepromAcc.Buffer))? 0 : -EIO;

      // update chached EEPROM
      if (argcmd.EepromAcc.MaxSize > PD_EEPROM_SIZE)
         argcmd.EepromAcc.MaxSize = PD_EEPROM_SIZE;

      memcpy(pd_board[board].Eeprom.u.WordValues,
             argcmd.EepromAcc.Buffer,
             argcmd.EepromAcc.MaxSize*sizeof(u16));

      break;

   case  IOCTL_PWRDAQ_BRDENABLEINTERRUPT:
      retf = (pd_adapter_enable_interrupt(board, argcmd.dwParam[0]) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_BRDTESTINTERRUPT:
      retf = (pd_adapter_test_interrupt(board) ? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_BRDSTATUS:
      retf = (pd_adapter_get_board_status(board, &argcmd.Event)?0:-EIO);
      break;

   case  IOCTL_PWRDAQ_BRDSETEVNTS1:
      retf = (pd_adapter_set_board_event1(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_BRDSETEVNTS2:
      retf = (pd_adapter_set_board_event2(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_BRDFWLOAD: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_BRDFWEXEC: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_BRDREGWR:
      retf = (pd_dsp_reg_write(board, argcmd.dwParam[0],
                               argcmd.dwParam[1]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_BRDREGRD:
      retf = (pd_dsp_reg_read(board, argcmd.dwParam[0],
                              (u32*)&argcmd.dwParam[1]))? 0 : -EIO;
      break;

      // AI
   case  IOCTL_PWRDAQ_AISETCFG:
      retf = (pd_ain_set_config(board, argcmd.dwParam[0],
                                argcmd.dwParam[1],argcmd.dwParam[2]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISETCVCLK:
      retf = (pd_ain_set_cv_clock(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISETCLCLK:
      retf = (pd_ain_set_cl_clock(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISETCHLIST:
      retf = (pd_ain_set_channel_list(board, argcmd.AinSyncCfg.dwChListSize,
                                      argcmd.AinSyncCfg.dwChList))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISETEVNT:
      retf = (pd_ain_set_events(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISTATUS:
      retf = (pd_ain_get_status(board, (u32*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AICVEN:
      retf = (pd_ain_set_enable_conversion(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISTARTTRIG:
      retf = (pd_ain_sw_start_trigger(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISTOPTRIG:
      retf = (pd_ain_sw_stop_trigger(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISWCVSTART:
      retf = (pd_ain_sw_cv_start(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AISWCLSTART:
      retf = (pd_ain_sw_cl_start(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AICLRESET:
      retf = (pd_ain_reset_cl(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AICLRDATA:
      retf = (pd_ain_clear_data(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AIRESET:
      retf = (pd_ain_reset(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AIGETVALUE:
      retf = (pd_ain_get_value(board, (u16*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AIGETSAMPLES:
      if (argcmd.bufParam.size > pd_board[board].AinSS.XferBufValues)
      {
         retf = -EINVAL; break;
      }
      pd_ain_get_samples(board,
                         argcmd.bufParam.size,
                         (u16*)pd_board[board].AinSS.pXferBuf,
                         &ret);
      pd_copy_to_user16((uint16_t*)argcmd.bufParam.address,
                        (u16*)pd_board[board].AinSS.pXferBuf,
                        ret*sizeof(u16));
      retf = (ret < 0) ? -EIO : ret;
      break;

   case  IOCTL_PWRDAQ_AIGETXFERSAMPLES:
      ret = pd_ain_get_xfer_samples(board, (uint16_t *)pd_board[board].AinSS.pXferBuf);
      pd_copy_to_user16((uint16_t*)argcmd.bufParam.address,
		        (uint16_t*)pd_board[board].AinSS.pXferBuf,
			pd_board[board].AinSS.BlkXferValues);
      retf = (ret < 0) ? -EIO : ret;
      break;

   case  IOCTL_PWRDAQ_AISETXFERSIZE:
      retf = (pd_ain_set_xfer_size(board,
                                   argcmd.dwParam[0]
                                  ))? 0 : -EIO;
      break;


   case  IOCTL_PWRDAQ_AISETSSHGAIN: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_AIENABLECLCOUNT: retf = -ENOSYS;
      break;
   case  IOCTL_PWRDAQ_AIENABLETIMER: retf = -ENOSYS;
      break;
   case  IOCTL_PWRDAQ_AIFLUSHFIFO: retf = -ENOSYS;
      break;
   case  IOCTL_PWRDAQ_AIGETSAMPLECOUNT: 
      //retf = -ENOSYS;
      argcmd.dwParam[0] = pd_board[board].AinSS.BufInfo.Count;
      break;
      // AO
   case  IOCTL_PWRDAQ_AOSETCFG:
      retf = (pd_aout_set_config(board,
                                 argcmd.dwParam[0],
                                 argcmd.dwParam[1]
                                ))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSETCVCLK:
      retf = (pd_aout_set_cv_clk(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSETEVNT:
      retf = (pd_aout_set_events(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSTATUS:
      retf = (pd_aout_get_status(board, (u32*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOCVEN:
      retf = (pd_aout_set_enable_conversion(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSTARTTRIG:
      retf = (pd_aout_sw_start_trigger(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSTOPTRIG:
      retf = (pd_aout_sw_stop_trigger(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOSWCVSTART:
      retf = (pd_aout_sw_cv_start(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOCLRDATA:
      retf = (pd_aout_clear_data(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AORESET:
      retf = (pd_aout_reset(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOPUTVALUE:
      retf = (pd_aout_put_value(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_AOPUTBLOCK:
      if (argcmd.dwParam[0]>AOUT_BLKSIZE*2)
      {
         retf = -EINVAL; break;
      }
      pd_copy_from_user32((u32*)pd_board[board].AoutSS.pXferBuf,
                          (u32*)argcmd.bufParam.address,
                          argcmd.bufParam.size*sizeof(u32)
                         );
      retf = (pd_aout_put_block(board,
                                argcmd.bufParam.size,                //num values
                                (u32*)pd_board[board].AoutSS.pXferBuf, //pdwBuf
                                (u32*)&argcmd.bufParam.size           //pdwCount
                               ))? 0 : -EIO;
      break;

      // DIO
   case  IOCTL_PWRDAQ_DISETCFG:
      retf = (pd_din_set_config(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_DISTATUS:
      retf = (pd_din_status(board, (u32*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_DIREAD:
      retf = (pd_din_read_inputs(board, (u32*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_DICLRDATA:
      retf = (pd_din_clear_data(board)? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_DIRESET:
      retf = (pd_din_reset(board)? 0 : -EIO);
      break;

   case  IOCTL_PWRDAQ_DOWRITE:
      retf = (pd_dout_write_outputs(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_DORESET:
      retf = (pd_dout_write_outputs(board, 0))? 0 : -EIO; // reset writes all zeroes
      break;

   case  IOCTL_PWRDAQ_DIO256CMDWR:
      retf = (pd_dio256_write_output(board,
                                     argcmd.dwParam[0],
                                     argcmd.dwParam[1]
                                    )
             )? 0 : -EIO;;
      break;

   case  IOCTL_PWRDAQ_DIO256CMDRD:
      retf = (pd_dio256_read_input(board,
                                   argcmd.dwParam[0],
                                   (u32*)&argcmd.dwParam[1]
                                  )
             )? 0 : -EIO;;
      break;

      //---------------------------------------------------------------
      // PdDIODMASet  -- write 3 param, DIO256
   case IOCTL_PWRDAQ_DIODMASET:
      retf = (pd_dio_dmaSet(board, argcmd.dwParam[0],
                            argcmd.dwParam[1], argcmd.dwParam[2])
             )? 0 : -EIO;
      break;

      //---------------------------------------------------------------
      // PdAODMASet  -- write 3 param, DIO256
   case IOCTL_PWRDAQ_AODMASET:
      retf = (pd_aout_dmaSet(board, argcmd.dwParam[0],
                             argcmd.dwParam[1], argcmd.dwParam[2])
             )? 0 : -EIO;
      break;


      //---------------------------------------------------------------
      // PdDIO256SetIntrMask  -- write 2 param to DSP register
   case IOCTL_PWRDAQ_DIO256SETINTRMASK:
      pd_copy_from_user32((u32*)pd_board[board].DinSS.intrMask,
                          (u32*)argcmd.bufParam.address,
                          argcmd.bufParam.size*sizeof(u32));

      DPRINTK("DIO SetIntMask ");
      for (i=0; i<8; i++)
      {
         DPRINTK("reg%d:0x%x ", i, pd_board[board].DinSS.intrMask[i]);
      }
      DPRINTK("\n");

      retf = (pd_dio256_setIntrMask(board)) ? 0 : -EIO;
      break;

      //---------------------------------------------------------------
      // PdDIO256GetIntrData  -- write 2 param to DSP register
   case IOCTL_PWRDAQ_DIO256GETINTRDATA:
      retf = (pd_dio256_getIntrData(board))? 0 : -EIO;

      DPRINTK("DIO GetIntrData ");
      for (i=0; i<16; i++)
      {
         DPRINTK("reg%d:0x%x ", i, pd_board[board].DinSS.intrData[i]);
      }
      DPRINTK("\n");

      pd_copy_to_user32((u32*)argcmd.bufParam.address,
                        (u32*)pd_board[board].DinSS.intrData,
                        argcmd.bufParam.size*sizeof(u32));
      break;

      //---------------------------------------------------------------
      // PdDIO256IntrEnable -- write 2 param to DSP register
   case IOCTL_PWRDAQ_DIO256INTRREENABLE:
      DPRINTK("DIO Intr Enable %d\n", argcmd.dwParam[0]);
      retf = (pd_dio256_intrEnable(board, argcmd.dwParam[0]))? 0 : -EIO;;
      break;


      // UCT
   case  IOCTL_PWRDAQ_UCTSETCFG:
      retf = (pd_uct_set_config(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTSTATUS: retf = -ENOSYS;
      retf = (pd_uct_get_status(board, (u32*)&argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTWRITE: retf = -ENOSYS;
      retf = (pd_uct_write(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTREAD: retf = -ENOSYS;
      retf = (pd_uct_read(board,
                          argcmd.dwParam[0],
                          (u32*)&argcmd.dwParam[1]
                         )
             )? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTSWGATE: retf = -ENOSYS;
      retf = (pd_uct_set_sw_gate(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTSWCLK: retf = -ENOSYS;
      retf = (pd_uct_sw_strobe(board))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_UCTRESET: retf = -ENOSYS;
      retf = (pd_uct_reset(board))? 0 : -EIO;
      break;

      //CALDIAG
   case  IOCTL_PWRDAQ_CALSETCFG: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_CALDACWRITE:
      retf = (pd_cal_dac_write(board, argcmd.dwParam[0]))? 0 : -EIO;
      break;

   case  IOCTL_PWRDAQ_DIAGPCIECHO: retf = -ENOSYS;
      break;

   case  IOCTL_PWRDAQ_DIAGPCIINT: retf = -ENOSYS;
      break;

      // test only ioctls
   case IOCTL_PWRDAQ_TESTEVENTS:
      DPRINTK_I("test interrupt from dispatch routine\n");
      pd_adapter_enable_interrupt(board, 1);
      retf = (pd_adapter_test_interrupt(board))? 0 : -EIO;
      pd_adapter_enable_interrupt(board, 0);
      break;

   case IOCTL_PWRDAQ_TESTBUFFERING:
      //test_pd_ain_buffering(board);
      retf = -ENOSYS; // 0
      break;

   case IOCTL_PWRDAQ_GETADCFIFOSIZE:
      argcmd.dwParam[0] = 
      pd_board[board].Eeprom.u.Header.ADCFifoSize * 0x400;

      DPRINTK_I("pd_ioctl: ADC FIFO size =%d\n", argcmd.dwParam[0]);
      retf=0;
      break;

   case IOCTL_PWRDAQ_GETSERIALNUMBER:
      pd_copy_to_user8((u8*)(argcmd.bufParam.address), (u8*)&pd_board[board].Eeprom.u.Header.SerialNumber, PD_SERIALNUMBER_SIZE);
      retf=0;
      break;

   case IOCTL_PWRDAQ_GETMANFCTRDATE:
      pd_copy_to_user8((u8*)(argcmd.bufParam.address), (u8*)&pd_board[board].Eeprom.u.Header.ManufactureDate, PD_DATE_SIZE);
      retf=0;
      break;

   case IOCTL_PWRDAQ_GETCALIBRATIONDATE:
      pd_copy_to_user8((u8*)(argcmd.bufParam.address), (u8*)&pd_board[board].Eeprom.u.Header.CalibrationDate, PD_DATE_SIZE);
      retf=0;
      break;

#ifdef MEASURE_LATENCY
   case IOCTL_PWRDAQ_MEASURE_LATENCY:
      argcmd.dwParam[0] = pd_board[board].tscLow;
      argcmd.dwParam[1] = pd_board[board].tscHigh;
      retf=0;
      break;
#endif

   } // switch

   _fw_spinunlock

   if((void*)arg != NULL)
      pd_copy_to_user32((u32*)arg, (u32*)&argcmd, sizeof(TCmd));

   return retf;
}


///////////////////////////////////////////////////////////////////////
//
//       Name:  pd_fasync
//
//   Function:  This gets called when the user-space process changes the
//              FASYNC flag.
//
//  Arguments:  Hmmm...
//
//    Returns:  0 if everything went well, and "what errno should be" if
//              there's a problem.
//
int pd_fasync(
             int fd,
             struct file *file,
             int mode
             ) {
   int real_minor, board, board_minor, ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   ret = fasync_helper(fd, file, mode, &pd_board[board].fasync);

   DPRINTK_I("board %d, %s (minor %u) got fasync mode %d... ret=%d fd=%d\n", board,
             pd_devices_by_minor[board_minor], real_minor, mode, ret, fd);

   return ret;
}


///////////////////////////////////////////////////////////////////////
//
//       Name:  pd_proc_read
//
//   Function:  This gets called when the user-space process accesses
//              /proc/powerdaq file
//
//  Arguments:  Defined in proc_fs.h
//
#define LIMITPAGE (PAGE_SIZE-80)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
static int pd_proc_read (char *buf, char **start, off_t offset, int len, int unused)
#else
static int pd_get_info (char *buf, char **start, off_t offset, int len)
#endif
{
   int i;
   char *modelname;
   u32 t, id;

   // Let's print information about PD boards installed
   i = 0; len = 0;
   len += sprintf(buf+len, "PowerDAQ Driver, version %d.%d.%d\n\n", PD_VERSION_MAJOR, PD_VERSION_MINOR, PD_VERSION_EXTRA);
   len += sprintf(buf+len, "PowerDAQ Info: %d board(s) installed\n\n", num_pd_boards);
   for (i = 0; i < num_pd_boards; i++)
   {
      if(PD_IS_PDXI(pd_board[i].PCI_Config.SubsystemID))
      {
         modelname = pdxi_models[pd_board[i].PCI_Config.SubsystemID - 0x211].modelname;
         id = pd_board[i].PCI_Config.SubsystemID - 0x100;
      }
      else
      {
         modelname = pd_models[pd_board[i].PCI_Config.SubsystemID - 0x101].modelname;
         id = pd_board[i].PCI_Config.SubsystemID;
      }

      // print board name and s/n
      len+= sprintf(buf+len, "PowerDAQ board #%d type:\t%18s\n\ts/n:\t%8s\n",
                    i+1,
                    modelname,
                    pd_board[i].Eeprom.u.Header.SerialNumber);
      if (len > LIMITPAGE) return len;

      // print ADC FIFO size
      if(PD_IS_DIO(id) || PD_IS_MFX(id) || PDL_IS_DIO(id) || PDL_IS_MFX(id))
      {
          len+= sprintf(buf+len, "\tInput Fifo size:\t%d samples\n\tCL FIFO Size:\t%d entries\n",
                        pd_board[i].Eeprom.u.Header.ADCFifoSize*1024,
                        pd_board[i].Eeprom.u.Header.CLFifoSize*256);
          if (len > LIMITPAGE) return len;
      }

      // print DAC FIFO size
      len+= sprintf(buf+len, "\tOutput Fifo size:\t%d samples\n",
                    pd_board[i].Eeprom.u.Header.DACFifoSize*1024);
      if (len > LIMITPAGE) return len;


      // print mfg and cal dates
      len+= sprintf(buf+len, "\tMfg. date:\t%12s\n\tCal. date:\t%12s\n",
                    pd_board[i].Eeprom.u.Header.ManufactureDate,
                    pd_board[i].Eeprom.u.Header.CalibrationDate);
      if (len > LIMITPAGE) return len;

      // print base address and IRQ
      len+= sprintf(buf+len, "\tBase address:\t0x%lx\n\tIRQ line:\t0x%x\n",
                    pd_board[i].PCI_Config.BaseAddress0,
                    pd_board[i].PCI_Config.InterruptLine);
      if (len > LIMITPAGE) return len;

      // print DSP revision
      len+= sprintf(buf+len, "\tDSP: v%d\n", (pd_board[i].fwTimestamp[0]&0x0F0000)>>16);
      if (len > LIMITPAGE) return len;

      // print firmware revision
      t = pd_board[i].fwTimestamp[0] >> 20;
      len+= sprintf(buf+len, "\tFirmware type:\t%s rev: %d.%d/%x\n",
                    (t<1)?"Generic":((t<2)?"AO":((t<3)?"DIO":((t<4)?"LMF":((t<5)?"MFx":"Unknown")))),
                    (pd_board[i].fwTimestamp[0]&0xF00)>>8, (pd_board[i].fwTimestamp[0]&0xFF), 
                    pd_board[i].fwTimestamp[1]);
      if (len > LIMITPAGE) return len;

      // print xfer mode
      len+= sprintf(buf+len, "\tTransfer mode:\t%d\n\n",
                    pd_board[i].dwXFerMode);
      if (len > LIMITPAGE) return len;
   }
   return len;
}

///////////////////////////////////////////////////////////////////////////

#ifdef ALLOCMMAPBUF
int pd_mmap(struct file *file, struct vm_area_struct *vma)
{
   int real_minor, board, board_minor;
   int ret;
   void* buf;
   TBuf_Info  *pDaqBuf; 

   // find out what are we talking about
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
   real_minor = iminor(file->f_dentry->d_inode);
#else
   real_minor = MINOR(file->f_dentry->d_inode->i_rdev);
#endif
   board = real_minor / PD_MINOR_RANGE;
   board_minor = real_minor % PD_MINOR_RANGE;

   DPRINTK_I("trying to mmap board %d, %s (minor %u)\n",
             board, pd_devices_by_minor[board_minor], real_minor);

   // now let's check - was ain or aout buffer allocated
   switch (board_minor)
   {
   case PD_MINOR_AIN:
   case PD_MINOR_DIN:
   case PD_MINOR_UCT:
      if (!pd_board[board].AinSS.BufInfo.databuf)
         return -EIO;

      pDaqBuf = &pd_board[board].AinSS.BufInfo;
      break;

   case PD_MINOR_AOUT:
   case PD_MINOR_DOUT:
      if (!pd_board[board].AoutSS.BufInfo.databuf)
         return -EIO;

      pDaqBuf = &pd_board[board].AoutSS.BufInfo;
      break;
   default:
      return -ENOSYS;
   }

   if (!pDaqBuf)
      return -EFAULT;

   buf = pDaqBuf->databuf;

   // map it!
   if ((ret = rvmmap(buf, pDaqBuf->BufSizeInBytes, vma)) < 0)
   {
      DPRINTK_F("rvmmap fails with %d\n", ret);
      return ret;
   } else
   {
      DPRINTK_F("rvmmap succeed with %d\n", ret);
      return 0;
   }
}
#endif

///////////////////////////////////////////////////////////////////////////
struct file_operations pd_fops = 
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)

//   owner:	THIS_MODULE,
#endif
   .llseek =  NULL,     
   .read =    pd_read,  
   .write =   pd_write, 
   .ioctl =   pd_ioctl, 
#ifdef ALLOCMMAPBUF
   .mmap =    pd_mmap,  
#else
   .mmap =    NULL,     
#endif
   .open =    pd_open,  
   .flush =   NULL, 
   .release = pd_release, 
   .fsync =   NULL,
   .fasync =  pd_fasync,  
};

//////////////////////////////////////////////////////////////////////////
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
static struct proc_dir_entry pd_proc_entry = 
{
   0,                            /* low_ino    */
   sizeof("pwrdaq")-1,           /* namelen    */
   "pwrdaq",                     /* name       */
   S_IFREG | S_IRUGO,            /* mode       */
   1,                            /* nlink      */
   0,                            /* uid        */
   0,                            /* gid        */
   0,                            /* size       */
   NULL,                         /* ops        */
   pd_proc_read,                 /* get_info   */
   NULL,                         /* fill_inode */
   NULL,                         /* next       */
   NULL,                         /* parent     */
   NULL,                         /* subdir     */
   NULL                          /* data       */
};
#endif

//////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
//
//       name:  init_module
//
//   function:  Initialize the hardware and the device driver.
//
//  arguments:  None.
//
//    returns:  0 if the module loaded successfully, and 1 if not.
//
static int __init powerdaq_init(void) 
{
   struct pci_dev *dev = NULL;
   int i, ret;
   
   printk("PowerDAQ Driver, Copyright (C) 2000,2004 United Electronic Industries, Inc.\n");

   DPRINTK_N(PD_VERSION);
   num_pd_boards = 0;

   while ((dev = pci_find_device(MOTOROLA_VENDORID, DSP56301_DEVICEID, dev)))
   {
      deal_with_device(dev);
   }

   if (num_pd_boards == 0)
   {
      printk("No PowerDAQ board was found\n");
      return -ENODEV;
   }

   // register the device with the kernel
   DPRINTK_N("registering driver, major number %d\n>\n>\n", pd_major);
   if ((ret = register_chrdev(pd_major, "powerdaq", &pd_fops)))
   {
      DPRINTK_F("unable to get major number %d\n", pd_major);
      return -EBUSY;
   }

   // initialize /proc/powerdaq file
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
   proc_register(&proc_root, &pd_proc_entry);
#else
   create_proc_info_entry("pwrdaq", 0, NULL, pd_get_info);
#endif

// --------- test code --------
   init_waitqueue_head(&whead);
// --------- test code ends ---

   // prepare the bottom half task
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   if(rqstirq)
   {
      for (i = 0; i < num_pd_boards; i++)
      {
         // setup interrupt handler thread
         ret = pthread_create(&rt_isr_thread[i], NULL, pd_rt_bottom_half, (void*)(long)i);
         if (ret != 0)
         {
            DPRINTK("powerdaq_init: can't create interrupt thread\n");
            return ret;
         }
   
      }
   }
   rtl_allow_interrupts();
#elif defined(_PD_RTAI)
   if(rqstirq)
   {
      for (i = 0; i < num_pd_boards; i++)
      {
         rt_task_init (&rt_bh_task[i], pd_rt_bottom_half, i, 10000, 2, 1, 0);
   	     rt_sem_init (&rt_bh_sem[i], 0);
         rt_task_resume (&rt_bh_task[i]);
      }
   }
#else
   if(rqstirq)
   {
      for (i = 0; i < PD_MAX_BOARDS; i++)
      {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
         INIT_WORK(&pd_bh_task[i], pd_bottom_half, (void*)(unsigned long)i);
#else
         pd_bh_task[i].routine = pd_bottom_half;
#endif
      }
   }
#endif

   return 0;
}

//////////////////////////////////////////////////////////////////////////
//
//       name:  cleanup_module
//
//   function:  Shut down the hardware and the device driver.
//
//  arguments:  None.
//
//    returns:  Nothing.
//
static void __exit powerdaq_cleanup(void) 
{
   int i;
   AllocContigMem Mem;
   
   DPRINTK_N("unloading PowerDAQ driver...\n");

   for (i = 0; i < num_pd_boards; i ++)
   {
      DPRINTK_N("shutting down card %d\n", i);

      pd_reset_dsp(i);
            
      //-------------------------------------------------------------------
      // Release memory chunks for bus-mastering operations
      Mem.idx = AI_PAGE0;
      pd_dealloc_contig_memory(i, &Mem);

      Mem.idx = AI_PAGE1;
      pd_dealloc_contig_memory(i, &Mem);

      Mem.idx = AO_PAGE0;
      pd_dealloc_contig_memory(i, &Mem);

#if defined(_PD_RTL) || defined(_PD_RTLPRO)

      // Destroy synchronization objects
      pthread_cond_destroy(&pd_board[i].AinSS.synch->event);
      pthread_mutex_destroy(&pd_board[i].AinSS.synch->event_lock);
      pthread_cond_destroy(&pd_board[i].AoutSS.synch->event);
      pthread_mutex_destroy(&pd_board[i].AoutSS.synch->event_lock);
      pthread_cond_destroy(&pd_board[i].DinSS.synch->event);
      pthread_mutex_destroy(&pd_board[i].DinSS.synch->event_lock);
      pthread_cond_destroy(&pd_board[i].DoutSS.synch->event);
      pthread_mutex_destroy(&pd_board[i].DoutSS.synch->event_lock);
      pthread_cond_destroy(&pd_board[i].UctSS.synch->event);
      pthread_mutex_destroy(&pd_board[i].UctSS.synch->event_lock);
#elif defined(_PD_RTAI)
      // Destroy synchronization objects
      rt_cond_destroy(&pd_board[i].AinSS.synch->event);
      rt_mutex_destroy(&pd_board[i].AinSS.synch->event_lock);
      rt_cond_destroy(&pd_board[i].AoutSS.synch->event);
      rt_mutex_destroy(&pd_board[i].AoutSS.synch->event_lock);
      rt_cond_destroy(&pd_board[i].DinSS.synch->event);
      rt_mutex_destroy(&pd_board[i].DinSS.synch->event_lock);
      rt_cond_destroy(&pd_board[i].DoutSS.synch->event);
      rt_mutex_destroy(&pd_board[i].DoutSS.synch->event_lock);
      rt_cond_destroy(&pd_board[i].UctSS.synch->event);
      rt_mutex_destroy(&pd_board[i].UctSS.synch->event_lock);
#endif

      // free up the memory allocated for the synch objects
      kfree(pd_board[i].AinSS.synch);
      kfree(pd_board[i].AoutSS.synch);
      kfree(pd_board[i].DinSS.synch);
      kfree(pd_board[i].DoutSS.synch);
      kfree(pd_board[i].UctSS.synch);

      // free up the remapped pages
      iounmap(pd_board[i].address);

      // free the IRQ
      if(rqstirq)
      {
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
        // Kill the isr thread
        pthread_cancel(rt_isr_thread[i]);
        pthread_join (rt_isr_thread[i], NULL);

        // Destroy synchronization objects
        pthread_mutex_destroy(&pd_board_ext[i].user_isr_lock);

        // free irq
        rtl_hard_disable_irq(pd_board[i].irq);
        rtl_free_irq(pd_board[i].irq);
#elif defined(_PD_RTAI)
        rt_shutdown_irq(pd_board[num_pd_boards].irq);
        rt_disable_irq(pd_board[num_pd_boards].irq);
        rt_free_global_irq(pd_board[num_pd_boards].irq);

        rt_task_delete (&rt_bh_task[i]);
	    rt_sem_delete (&rt_bh_sem[i]);
#else
        free_irq(pd_board[i].irq, (void *)&pd_board[i]);
#endif
      }
   }

   // stop /proc/powerdaq file operations
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
   proc_unregister(&proc_root, pd_proc_entry.low_ino);
#else
   remove_proc_entry("pwrdaq", NULL);
#endif


   DPRINTK_N("unregistering device\n");
   if (unregister_chrdev(pd_major, "powerdaq"))
   {
      DPRINTK_F("device unregistering failed\n");
   } 
   else
   {
      DPRINTK_N("PowerDAQ driver unloaded\n>\n>\n>\n");
   }
}

//////////////////////////////////////////////////////////////////////////
// Event notification functions
//
// This function puts process asleep
//
int pd_sleep_on_event(int board, PD_SUBSYSTEM ss, int event, int timeoutms) 
{
   int tret, everet;
   TSynchSS *synch;

   tret = everet = 0;
   
   DPRINTK_N("pd_sleep_on_event: cur:%d brd:%d ss:%d evt:%x\n",
             current->pid, board, ss, event);

   switch (ss)
   {
   case AnalogIn:
      synch = pd_board[board].AinSS.synch; 
      break;

   case AnalogOut:
      synch = pd_board[board].AoutSS.synch;
      break;

   case DigitalIn:
      if (event & eDInEvent)
      {
         synch = pd_board[board].DinSS.synch;
      } 
      else
      {
         synch = pd_board[board].AinSS.synch;
      }
      break;

   case DigitalOut:
      synch = pd_board[board].AoutSS.synch;
      break;

   case CounterTimer:
   case DSPCounter:
      if (event & (eUct0Event | eUct1Event | eUct2Event))
      {
         synch = pd_board[board].UctSS.synch;
      }
      else
      {
         synch = pd_board[board].AinSS.synch;
      }
      break;
   case CalDiag:
   case BoardLevel:
   default:
      everet = 0;
      tret = -1;
      synch = NULL;
      break;
   }

   if(NULL != synch)
   {
      synch->wakeupEvents = event;
      tret = pd_wait_for_event(board, synch, timeoutms);
      everet = synch->notifiedEvents;
      synch->notifiedEvents = 0;
   }

   DPRINTK_N("tret=%d %s\n", tret, ((tret <= 0)?"TIMEOUT":""));
   if (tret <= 0) 
       everet |= eTimeout;
   return everet;
}

//
// This function awakes a process
//
int pd_notify_event(int board, PD_SUBSYSTEM ss, int event) 
{
   DPRINTK_N("wake up!: cur:%d brd:%d ss:%d evt:%x\n",
             current->pid, board, ss, event);

   switch (ss)
   {
   case AnalogIn:
      pd_board[board].AinSS.synch->notifiedEvents = event;
      pd_set_event(board, pd_board[board].AinSS.synch); 
      break;

   case AnalogOut:
      pd_board[board].AoutSS.synch->notifiedEvents = event;
      pd_set_event(board, pd_board[board].AoutSS.synch);
      break;

   case DigitalIn:
      pd_board[board].DinSS.synch->notifiedEvents = event;
      pd_set_event(board, pd_board[board].DinSS.synch);
      break;

   case DigitalOut:
      pd_board[board].DoutSS.synch->notifiedEvents = event;
      pd_set_event(board, pd_board[board].DoutSS.synch);
      break;

   case CounterTimer:
   case DSPCounter:
      pd_board[board].UctSS.synch->notifiedEvents = event;
      pd_set_event(board, pd_board[board].UctSS.synch);
      break;

   case CalDiag:
   case BoardLevel:
   default:
      break;
   }

   return event;
}


//////////////////////////////////////////////////////////////////////////
// Kernel functions wrapper
//
// This function wrapper needed to allow library to stay the same while
// you might want to recompile driver for different kernels/RT-Linux and RTAI
//
//--------------------------------------------------------------------
void pd_kill_fasync(struct fasync_struct *fa, int sig)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)) || defined(PD_USING_REDHAT_62)
   kill_fasync(&fa, sig, 0);
#else
   kill_fasync(fa, sig);
#endif
}
//--------------------------------------------------------------------
unsigned long pd_copy_from_user32(u32* to, u32* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_from_user32(u32* to, u32* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd_copy_to_user32(u32* to, u32* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_to_user32(u32* to, u32* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

//--------------------------------------------------------------------
unsigned long pd_copy_from_user16(u16* to, u16* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_from_user16(u16* to, u16* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd_copy_to_user16(u16* to, u16* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd_copy_to_userbuf16(u16* to, u16* from, u32 len)
{
#ifdef _COPYSAMPLESTOUSERBUFFER
   return copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_to_user16(u16* to, u16* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

//--------------------------------------------------------------------
unsigned long pd_copy_from_user8(u8* to, u8* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_from_user8(u8* to, u8* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_from_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd_copy_to_user8(u8* to, u8* from, u32 len)
{
#ifndef _NO_USERSPACE
   return copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

unsigned long pd___copy_to_user8(u8* to, u8* from, u32 len)
{
#ifndef _NO_USERSPACE
   return __copy_to_user(to, from, len);
#else
   return(unsigned long)memcpy(to, from, len);
#endif
}

//--------------------------------------------------------------------
int pd_get_user32(u32* value, u32* address)
{
   u32 ret;
   u32 err;
   err = get_user(ret, address);
   *value = ret;
   return err;
}

int pd___get_user32(u32* value, u32* address)
{
   u32 ret;
   u32 err;
   err = get_user(ret, address);
   *value = ret;
   return err;
}

int pd_get_user16(u16* value, u16* address)
{
   u16 ret;
   u32 err;
   err = get_user(ret, address);
   *value = ret;
   return err;
}

int pd___get_user16(u16* value, u16* address)
{
   u16 ret;
   u32 err;
   err = get_user(ret, address);
   *value = ret;
   return err;
}

//--------------------------------------------------------------------
void pd_udelay(u32 usecs)
{
   udelay(usecs);
}

void pd_mdelay(u32 msecs)
{
   mdelay(msecs);
}

//-------------------------------------------------------------------
void* pd_alloc_bigbuf(u32 size)
{
#ifdef ALLOCMMAPBUF
   return rvmalloc(size);
#else
   return kmalloc(size, GFP_KERNEL);
#endif
}

void pd_free_bigbuf(void* mem, u32 size)
{
#ifdef ALLOCMMAPBUF
   rvfree(mem, size);
#else
   kfree(mem);
#endif
}

//--------------------------------------------------------------------
void* pd_kmalloc(u32 size, u32 priority)
{
   return kmalloc(size, priority);
}

void pd_kfree(void* adr)
{
   kfree(adr);
}

//--------------------------------------------------------------------
unsigned int pd_readl(void *address)
{
   return readl(address);
}


void pd_writel(unsigned int value, void *address)
{
   writel(value, address);
}


//--------------------------------------------------------------------
int pd_wait_for_event(int board, TSynchSS *synch, int timeoutms)
{
   int tret;

   DPRINTK("pd_wait_for_event: board %d, timeout %d\n", board, timeoutms);

   // unlock the spin lock so that we don't deadlock when the event occurs
   _fw_spinunlock

#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   struct timespec timeout;

   // Get the lock, pthread_cond_wait will unlock it while waiting and
   // relock it when it gets signaled
   tret = pthread_mutex_lock(&synch->event_lock);
   //tret = pthread_mutex_trylock(&synch->event_lock);
   if (tret != 0)
   {
       DPRINTK_N("pd_wait_for_event: Could not acquire event lock\n");
       _fw_spinlock
       return tret;
   }

   DPRINTK_N("mutex locked\n");

   // Set a timeout of 5 sec from now for waiting for the answer of a module
   timeout.tv_sec =  time(0) + timeoutms/1000;
   timeout.tv_nsec = (timeoutms % 1000) + 1000000;

   tret = pthread_cond_timedwait(&synch->event, &synch->event_lock, &timeout);
   if (tret != 0)
   {
      if (tret == ETIMEDOUT)
      {
         DPRINTK_N("Timeout\n");
         tret = -1;
      } 
      else
      {
         DPRINTK_N("Unexpected error\n");
      }
   }
   else
      tret = 1;

   DPRINTK("cond Wait returned %d\n", tret);

   pthread_mutex_unlock(&synch->event_lock);

   DPRINTK("Mutex unlocked\n");
#elif defined (_PD_RTAI)
   RTIME rt_timeout;
   RTIME timeout_ns;

   // Get the lock, pthrea_cond_wait will unlock it while waiting and
   // relock it when it gets signaled
   tret = rt_mutex_lock(&synch->event_lock);
   if (tret != 0)
   {
       DPRINTK_N("pd_wait_for_event: Could not acquire event lock\n");
       _fw_spinlock
       return tret;
   }

   // Set a timeout of 5 sec from now for waiting for the answer of a module
   timeout_ns = timeoutms * 1000 * 1000;
   rt_timeout = nano2count(timeout_ns);

   tret = rt_cond_wait_timed(&synch->event, &synch->event_lock, rt_timeout);
   if (tret != 0)
   {
      if (tret == ETIMEDOUT)
      {
         DPRINTK_N("Timeout\n");
         tret = -1;
      } 
      else
      {
         DPRINTK_N("Unexpected error\n");
      }
   }
   else
      tret = 1; 

   rt_mutex_unlock(&synch->event_lock);
#else
   int toutjiffies = (timeoutms * HZ +999)/1000 + 1;
   tret = toutjiffies;
   wait_queue_t wait;
   init_waitqueue_entry(&wait, current);

   DPRINTK("pd_wait_for_event: about to go to sleep\n");
  
   // Handle the sleep on process ourself to avoid
   // race conditions that may occurr with interruptible_sleep_on_tmeout
   add_wait_queue(&synch->wait_q, &wait);
   while (tret > 0 && !signal_pending(current)) 
   {
       set_current_state(TASK_INTERRUPTIBLE);
       if (synch->notifiedEvents != 0) 
          break;
       tret = schedule_timeout(tret);

       DPRINTK("schedule_timeout ret %d\n", tret);
   }
   set_current_state(TASK_RUNNING);
   remove_wait_queue(&synch->wait_q, &wait);

   //tret = interruptible_sleep_on_timeout(&synch->wait_q, toutjiffies);
#endif // _PD_RTL

   _fw_spinlock

   return tret; 
}

int pd_set_event(int board, TSynchSS *synch)
{
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
   // Signal the event
   pthread_cond_signal(&synch->event);
#elif defined(_PD_RTAI)
   rt_cond_signal(&synch->event);
#else
   wake_up_interruptible(&synch->wait_q);
#endif // _PD_RTL

   return 0;
}

// This function enumerates devices and return -1 when no device is found
int pd_enum_devices(int board, char* board_name, int board_name_size, 
                    char *serial_number, int serial_number_size)
{
   char *modelname;
   char *serial;
   
   if(PD_IS_PDXI(pd_board[board].PCI_Config.SubsystemID))
      modelname = pdxi_models[pd_board[board].PCI_Config.SubsystemID - 0x211].modelname;
   else
      modelname = pd_models[pd_board[board].PCI_Config.SubsystemID - 0x101].modelname;
     
   serial = pd_board[board].Eeprom.u.Header.SerialNumber;


   if(board >= num_pd_boards)
   {
      return -1;
   }

   if(strlen(modelname) >= board_name_size)
      strncpy(board_name, modelname, board_name_size - 1);
   else
      strcpy(board_name, modelname);

   if(strlen(serial) >= serial_number_size)
      strncpy(serial_number, serial, serial_number_size - 1);
   else
      strcpy(serial_number, serial);

   return 0;
}

#if defined(_PD_RTL) || defined(_PD_RTLPRO)
int pd_register_user_isr(int board, TUser_isr user_isr, void* user_param)
{
   pthread_mutex_lock(&pd_board_ext[board].user_isr_lock);
   pd_board_ext[board].user_isr = user_isr;
   pd_board_ext[board].user_param = user_param;
   pthread_mutex_unlock(&pd_board_ext[board].user_isr_lock);

   return 0;
}


int pd_unregister_user_isr(int board)
{
   pthread_mutex_lock(&pd_board_ext[board].user_isr_lock);
   pd_board_ext[board].user_isr = NULL;
   pd_board_ext[board].user_param = NULL;
   pthread_mutex_unlock(&pd_board_ext[board].user_isr_lock);

   return 0;
}
#elif defined (_PD_RTAI)
int pd_register_user_isr(int board, TUser_isr user_isr, void* user_param)
{
   rt_mutex_lock(&pd_board_ext[board].user_isr_lock);
   pd_board_ext[board].user_isr = user_isr;
   pd_board_ext[board].user_param = user_param;
   rt_mutex_unlock(&pd_board_ext[board].user_isr_lock);

   return 0;
}


int pd_unregister_user_isr(int board)
{
   rt_mutex_lock(&pd_board_ext[board].user_isr_lock);
   pd_board_ext[board].user_isr = NULL;
   pd_board_ext[board].user_param = NULL;
   rt_mutex_unlock(&pd_board_ext[board].user_isr_lock);

   return 0;
}
#endif


//////////////////////////////////////////////////////////////////////////
//
// Routine Description:
//
//    Allocates non-paged pool memory for use by the Pd device.  Actual Input
//    and Output comes from the IRP system buffer, and are AllocMemoryIn and
//    AllocMemoryOut structures (see Pddrv.h).
//    This routine keeps track of memory allocations by storing allocation
//    information in the device extension.  If memory has alrady been allocated
//    for the Pd device, this routine will return an error.
//    If the entire allocation request can not be satisfied, then no allocation is
//    performed (actually, the part of the allocaion that succeeds is undone).
//    Callers to this funcion should also call PdUpdatePagetable().
//
//////////////////////////////////////////////////////////////////////////
int pd_alloc_contig_memory(int board, AllocContigMem *allocMemory) 
{
   int idx = allocMemory->idx;
   ULONG errors = FALSE;
   int status = 0;

   DPRINTK("Allocating %ld pages for index %d\n", allocMemory->size, allocMemory->idx);
   
   allocMemory->status = 0;

   if (idx & AIBM_GETPAGE) 
   {
       // return existing data
       idx &= 0x7;
       allocMemory->linearAddr = (unsigned long) pd_board[board].pLinBMB[idx];
       allocMemory->physAddr = (unsigned long) pd_board[board].DMAHandleBMB[idx];
       return status;
   }

   //
   // Check the device extension to see if memory has already
   // been allocated for this Pd device.  If so, return an
   // error
   //
   idx = allocMemory->idx & 0x7;
   if (pd_board[board].pSysBMB[idx]) 
   {
      allocMemory->status = 1;
      return 1;
   }

   pd_board[board].SizeBMB[idx] = allocMemory->size * PAGE_SIZE;
   //pd_board[board].pSysBMB[idx] = kmalloc(pd_board[board].SizeBMB[idx], GFP_KERNEL);
   pd_board[board].pSysBMB[idx] = pci_alloc_consistent(pd_board[board].dev, pd_board[board].SizeBMB[idx], &(pd_board[board].DMAHandleBMB[idx]));
   
   if (!pd_board[board].pSysBMB[idx]) 
   {
      pd_board[board].SizeBMB[idx] = 0;
      allocMemory->status = 1;
      status = ENOMEM;
      errors = TRUE;
      goto CLEANUP;
   }

   //
   // now get a physical address 
   //allocMemory->physAddr = virt_to_phys(pd_board[board].pSysBMB[idx]);
   //pd_board[board].pPhysBMB[idx] = (void *)allocMemory->physAddr;

   pd_board[board].pLinBMB[idx] = pd_board[board].pSysBMB[idx];
   allocMemory->linearAddr = (unsigned long) pd_board[board].pLinBMB[idx];

CLEANUP:
   
   return status;
}

//////////////////////////////////////////////////////////////////////////
//
// Routine Description:
//
//    Deallocates all memory that has been allocated for the Pd device.
//////////////////////////////////////////////////////////////////////////
void pd_dealloc_contig_memory(int board, AllocContigMem * pDeallocMemory)
{
   int idx = pDeallocMemory->idx;

   DPRINTK("Freeing memory for index %d\n", pDeallocMemory->idx);

   //
   // free the buffer
   //
   if (pd_board[board].pSysBMB[idx])
   {
      //kfree(pd_board[board].pSysBMB[idx]);
      pci_free_consistent(pd_board[board].dev, pd_board[board].SizeBMB[idx],
                          pd_board[board].pSysBMB[idx],
                          pd_board[board].DMAHandleBMB[idx]);
   
      pd_board[board].pSysBMB[idx] = NULL;
      //pd_board[board].pPhysBMB[idx] = 0;
      pd_board[board].pLinBMB[idx] = 0;
      pd_board[board].SizeBMB[idx] = 0;
   }
}


void pd_clean_device(int board)
{
#if defined(_PD_RTL) || defined(_PD_RTLPRO)
    pthread_mutex_unlock(&pd_board[board].AinSS.synch->event_lock);
    pthread_mutex_unlock(&pd_board[board].AoutSS.synch->event_lock);
    pthread_mutex_unlock(&pd_board[board].DinSS.synch->event_lock);
    pthread_mutex_unlock(&pd_board[board].DoutSS.synch->event_lock);
    pthread_mutex_unlock(&pd_board[board].UctSS.synch->event_lock);
#endif
}

module_init(powerdaq_init);
module_exit(powerdaq_cleanup);




